"""
Functions for validating user-provided arguments.

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import os
from parrot import process_input_data as pid
from parrot import encode_sequence

# Ensure that provided file is valid
def check_file_exists(f, name):
    file = os.path.abspath(f)
    if not os.path.isfile(file):
        err_str = f"{name} does not exist."
        raise FileNotFoundError(err_str)
    else:
        return file

def check_directory(dir, name):
    if not os.path.exists(dir):
        raise FileNotFoundError(f'{name} does not exist.')

# Returns the base filename prefix and the directory it is located in 
# from the absolute path
def split_file_and_directory(path):
    filename = path.split('/')[-1]
    directory = path[:-len(filename)]
    check_directory(directory, directory)
    base_filename = directory + filename.split('.')[0]
    return base_filename, directory

def set_encoding_scheme(encode_str):
    if encode_str == 'onehot':
        encoding_scheme = 'onehot'
        input_size = 20
        encoder = None
    elif encode_str == 'biophysics':
        encoding_scheme = 'biophysics'
        input_size = 9
        encoder = None
    else:
        encoding_scheme = 'user'
        encode_file = encode_str
        check_file_exists(encode_file, "Encoding file")
        encoder = encode_sequence.UserEncoder(encode_file)
        input_size = len(encoder)

    return encoding_scheme, encoder, input_size

def check_positive(arg, name):
    if arg < 1:
        raise ValueError(f'{name} must be a positive integer.')

def set_ml_task(n_classes, dtype):
    # Initialize network as classifier or regressor and set collate function
    if n_classes > 1:
        problem_type = 'classification'

        if dtype == 'sequence':
            collate_fnct = pid.seq_class_collate
        elif dtype == 'residues':
            collate_fnct = pid.res_class_collate
        else:
            raise ValueError('Invalid argument `--datatype`: must be "residues" or "sequence"')

    elif n_classes == 1:
        problem_type = 'regression'

        if dtype == 'sequence':
            collate_fnct = pid.seq_regress_collate
        elif dtype == 'residues':
            collate_fnct = pid.res_regress_collate
        else:
            raise ValueError('Invalid argument `--datatype`: must be "residues" or "sequence"')

    else:
        raise ValueError('Number of classes must be a positive integer.')

    return problem_type, collate_fnct

def check_between_zero_and_one(val, name):
    if val >= 1 or val <= 0:
        raise ValueError(f'{name} must be between 0 and 1.')

def check_stop_condition(stop_cond, n_epochs):
    if stop_cond == 'auto':
        if n_epochs > 10:
            print("Warning: Stop condition is set to 'auto' and num_epochs > 10." +
                   " Network training may take a long time.\n")
    elif stop_cond != 'iter':
        raise ValueError('Invalid argument for `--stop` -- must be "auto" or "iter".')




"""
Module with functions for processing an input datafile into a PyTorch-compatible
format.

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import math

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import Dataset

from parrot import encode_sequence, parrot_exceptions
from parrot.tools import dataset_warnings



# .................................................................
#
#
def read_tsv_raw(tsvfile, delimiter=None):
    """
    Internal function for parsing a tsv file. Ignores empty lines and
    allows for comment lines (lines that start with a # symbol). Does not
    do any other sanity checking, however. 
    Parameters
    ----------

    tsvfile : str
        Path to a whitespace-separated datafile

    delimiter : str or None
        Allows you to define the string to split columns in the file. Default
        is any whitespace character. Default = None (any whitespace).


    Returns
    -----------
    list
        Returns a list of strings where each element in the list 

    
    """

    # read in file
    with open(tsvfile) as fh:
        content = fh.readlines()

    # parse through based on delimiter. Note if delimiter=None then
    # this uses the default 
    lines = []    
    for line in content:     

        # skip empty lines
        if len(line.strip()) == 0:
            continue

        # skip comment lines
        if line.strip()[0] == '#':
            continue
        
        lines.append(line.strip().split(delimiter))

    return lines



# .................................................................
#
#
def __parse_lines(lines, datatype, validate=True):
    """
    Internal function for parsing a set of lines

    Parameters
    ----------
    lines : list
        A list of lists, where the sublists reflect the columns in a tsvfile. Should be the output
        from the read_tsv_raw() function.

    datatype : str
        Identifier that defines the type of data being passed in. Must be either 'residues', 'sequence'

    validate : bool
        If set to true, ensures the number of residue values equals the number of residues
 
    Returns
    -----------
    list
        Returns a parsed list of lists, where each sublist contains the structure
        [id, sequence, <data>]
        where <data> is either a single float (mode=sequence) or a set of floats 

    Raises
    ---------
    Exception
        If an error occurs while parsing the file, the linenumber of the file is printed as well as the
        idenity of the offending line.

    """

    
    # check the datatype is valid
    if datatype not in ['residues','sequence']:
        raise ValueError('Invalid datatype. Must be "residues" or "sequence".')
        
    # parse the lines
    try:

        data = []
        lc = 0

        # A value for each residue in a sequence
        if datatype == 'residues':	
            for x in lines:
                lc = lc + 1
                
                residue_data = np.array(x[2:], dtype=float)

                data.append([x[0], x[1], residue_data])

        # A single value per sequence
        elif datatype == 'sequence':  
            for x in lines:
                lc = lc + 1
                data.append([x[0], x[1], float(x[2])])

    except Exception as e:        
        print('Excecption raised on parsing input file...')
        print(e)
        print('')
        raise parrot_exceptions.IOExceptionParrot(f"Input data is not correctly formatted for datatype '{datatype}'.\nMake sure your datafile does not have empty lines at the end of the file.\nError on line {lc}:\n{x}")


    # if we want to validate each line 
    if validate:
        if datatype == 'residues':
            lc = 0
            for x in data:
                lc = lc + 1
                if len(x[1]) != len(x[2]):
                    raise parrot_exceptions.IOExceptionParrot(f"Input data is not correctly formatted for datatype '{datatype}'.\nInconsistent number of residue values and residues. Error on line {lc}:\n{x}")
                
        

    return data
    
    
# .................................................................
#
#
def parse_file(tsvfile, datatype, problem_type, num_classes, excludeSeqID=False, ignoreWarnings=False):
    """Parse a datafile containing sequences and values.

    Each line of of the input tsv file contains a sequence of amino acids, a value
    (or values) corresponding to that sequence, and an optional sequence ID. This 
    file will be parsed into a more convenient list of lists. 

    If excludeSeqID is False, then the format of each line in the file should be:
    <seqID> <sequence> <value(s)>

    If excludeSeqID is True, then the format of each line in the file should be:
    <sequence> <value(s)>

    `value(s)` will either be a single number if `datatype` is 'sequence' or a
    len(sequence) series of whitespace-separated numbers if it is 'residues'.

    If `problem_type` is 'regression', then each value can be any real number. But
    if it is 'classification' then each value should be an integer in the range
    [0-N] where N is the number of classes.

    Parameters
    ----------
    tsvfile : str
            Path to a whitespace-separated datafile
    datatype : str
            Description of the format of the values in `tsvfile`. Providing a string
            other than 'sequence' or 'residues' will produce unintended behavior.
    problem_type : str
            Description of the machine-learning task. Providing a string other than
            'regression' or 'classification' will produce unintended behavior.
    excludeSeqID : bool, optional
            Boolean indicating whether or not each line in `tsvfile` has a sequence ID
            (default is False)
    ignoreWarnings : bool, optional
            If False, assess the structure and balance of the provided dataset with 
            basic heuristics and display warnings for common issues.

    Returns
    -------
    list of lists
            A list representing the entire `tsvfile`. Each inner list corresponds to a
            single line in the file and has the format [seqID, sequence, values].
    """

    # read in and parse the TSV file. 
    lines  = read_tsv_raw(tsvfile)

    # Add a dummy seqID if none are provided
    if excludeSeqID:
        for line in lines:
            line.insert(0, '')

    data =  __parse_lines(lines, datatype)


    if not ignoreWarnings:
        # Check for identical sequences
        dataset_warnings.check_duplicate_sequences(data)

        # Check for class imbalance
        if problem_type == 'classification':
            dataset_warnings.check_class_imbalance(data)

        # Check for data distribution imbalance
        elif problem_type == 'regression':
            dataset_warnings.check_regression_imbalance(data)

    # if we're doing a classification problem...
    if problem_type == 'classification':

        # if a sequence classification
        if datatype == 'sequence':

            
            for sample in data:
                sample[2] = int(sample[2])

                # Validate that all of the class labels are valid
                if sample[2] >= num_classes or sample[2] < 0:
                    raise ValueError(f"Invalid class label on entry {sample[0]}.\nClass label was {sample[2]} but must be between 0 and {num_classes}")

        else:
            for sample in data:
                sample[2] = list(map(int, sample[2]))
                test = np.array(sample[2])

                if np.any(test < 0) or np.any(test >= num_classes):
                    raise ValueError(f"Invalid class label on entry {sample[0]}.\nClass labels must be between 0 and {num_classes}")


    return data


class SequenceDataset(Dataset):
    """A PyTorch-compatible dataset containing sequences and values

    Stores a collection of sequences as tensors along with their corresponding
    target values. This class is designed to be provided to PyTorch Dataloaders.


    Attributes
    ----------
    data : list of lists
            Each inner list represents a single sequence in the dataset and should
            have the format: [seqID, sequence, value(s)]
    encoding_scheme : str
            Description of how an amino acid sequence should be encoded as a numeric 
            vector. Providing a string other than 'onehot', 'biophysics', or 'user' 
            will produce unintended consequences.
    encoder: UserEncoder object, optional
            If encoding_scheme is 'user', encoder should be a UserEncoder object
            that can convert amino acid sequences to numeric vectors. If
            encoding_scheme is not 'user', use None.
    """

    def __init__(self, data, subset=np.array([]), encoding_scheme='onehot',
                 encoder=None):
        """
        Parameters
        ----------
        data : list of lists
                Each inner list represents a single sequence in the dataset and should
                have the format: [seqID, sequence, value(s)]
        subset : numpy int array, optional
                Array containing the indices of `data` that are to be a part of this
                dataset. Providing an empty array causes all of `data` to be integrated
                into the dataset (default is empty array).
        encoding_scheme : str
                Description of how an amino acid sequence should be encoded as a numeric 
                vector. Providing a string other than 'onehot', 'biophysics', or 'user' 
                will produce unintended consequences.
        encoder : UserEncoder object, optional
                If encoding_scheme is 'user', encoder should be a UserEncoder object
                that can convert amino acid sequences to numeric vectors. If
                encoding_scheme is not 'user', use None.
        """

        self.encoding_scheme = encoding_scheme
        self.encoder = encoder

        if len(subset) == 0:
            self.data = data
        else:
            all_data = data
            self.data = [all_data[x] for x in subset]

    def __len__(self):
        """Get the number of sequences in this dataset."""

        return len(self.data)

    def __getitem__(self, idx):
        """Get the sequence and values of a specified index.

        Converts a string sequence to an encoded numeric vector.

        Parameters
        ----------
        idx : int, optional
                Index of the desired sequence

        Returns
        -------
        tuple
                a tuple of a name, sequence vector and its corresponding values
        """

        if torch.is_tensor(idx):
            idx = idx.tolist()

        if self.encoding_scheme == 'onehot':
            sequence_vector = encode_sequence.one_hot(self.data[idx][1])
        elif self.encoding_scheme == 'biophysics':
            sequence_vector = encode_sequence.biophysics(self.data[idx][1])
        elif self.encoding_scheme == 'user':
            sequence_vector = self.encoder.encode(self.data[idx][1])

        name = self.data[idx][0]
        value = self.data[idx][2]
        sample = (name, sequence_vector, value)

        return sample


def seq_class_collate(batch):
    """Collates sequences and their values into a batch

    Transforms a collection of tuples of sequence vectors and values into a single
    tuple by stacking along a newly-created batch dimension. This function is 
    specifically designed for classification problems with sequence-mapped data.

    Parameters
    ----------
    batch : list
            A list of tuples of the form (sequence_vector, target_value(s))

    Returns
    -------
    tuple
            a tuple with concatenated names, sequence_vectors and target_values
    """

    names = [item[0] for item in batch]
    orig_seq_vectors = [item[1] for item in batch]
    orig_targets = [item[2] for item in batch]

    longest_seq = len(max(orig_seq_vectors, key=lambda x: len(x)))

    padded_seq_vectors = np.zeros([len(orig_seq_vectors), longest_seq, len(orig_seq_vectors[0][0])])

    for i, j in enumerate(orig_seq_vectors):
        padded_seq_vectors[i][0:len(j)] = j

    padded_seq_vectors = torch.FloatTensor(padded_seq_vectors)
    targets = torch.LongTensor(orig_targets)

    return (names, padded_seq_vectors, targets)


def seq_regress_collate(batch):
    """Collates sequences and their values into a batch

    Transforms a collection of tuples of sequence vectors and values into a single
    tuple by stacking along a newly-created batch dimension. This function is 
    specifically designed for regression problems with sequence-mapped data.

    Parameters
    ----------
    batch : list
            A list of tuples of the form (sequence_vector, target_value(s))

    Returns
    -------
    tuple
            a tuple with concatenated names, sequence_vectors and target_value
    """

    names = [item[0] for item in batch]
    orig_seq_vectors = [item[1] for item in batch]
    orig_targets = [[item[2]]for item in batch]

    longest_seq = len(max(orig_seq_vectors, key=lambda x: len(x)))

    padded_seq_vectors = np.zeros([len(orig_seq_vectors), longest_seq, len(orig_seq_vectors[0][0])])

    for i, j in enumerate(orig_seq_vectors):
        padded_seq_vectors[i][0:len(j)] = j

    padded_seq_vectors = torch.FloatTensor(padded_seq_vectors)
    targets = torch.FloatTensor(orig_targets)

    return (names, padded_seq_vectors, targets)


def res_class_collate(batch):
    """Collates sequences and their values into a batch

    Transforms a collection of tuples of sequence vectors and values into a single
    tuple by stacking along a newly-created batch dimension. This function is 
    specifically designed for classification problems with residue-mapped data. To
    account for sequences with different lengths, all sequence vectors are zero-
    padded to the length of the longest sequence in the batch

    Parameters
    ----------
    batch : list
            A list of tuples of the form (sequence_vector, target_value(s))

    Returns
    -------
    tuple
            a tuple with concatenated names, sequence_vectors and target_values
    """

    names = [item[0] for item in batch]
    orig_seq_vectors = [item[1] for item in batch]
    orig_targets = [item[2] for item in batch]

    longest_seq = len(max(orig_seq_vectors, key=lambda x: len(x)))

    padded_seq_vectors = np.zeros([len(orig_seq_vectors), longest_seq, len(orig_seq_vectors[0][0])])
    padded_targets = np.zeros([len(orig_targets), longest_seq])

    for i, j in enumerate(orig_seq_vectors):
        padded_seq_vectors[i][0:len(j)] = j

    for i, j in enumerate(orig_targets):
        padded_targets[i][0:len(j)] = j

    padded_seq_vectors = torch.FloatTensor(padded_seq_vectors)
    padded_targets = torch.LongTensor(padded_targets)

    return (names, padded_seq_vectors, padded_targets)


def res_regress_collate(batch):
    """Collates sequences and their values into a batch

    Transforms a collection of tuples of sequence vectors and values into a single
    tuple by stacking along a newly-created batch dimension. This function is 
    specifically designed for regression problems with residue-mapped data. To
    account for sequences with different lengths, all sequence vectors are zero-
    padded to the length of the longest sequence in the batch

    Parameters
    ----------
    batch : list
            A list of tuples of the form (sequence_vector, target_value(s))

    Returns
    -------
    tuple
            a tuple with concatenated names, sequence_vectors and target_values
    """

    names = [item[0] for item in batch]
    orig_seq_vectors = [item[1] for item in batch]
    orig_targets = [item[2] for item in batch]

    longest_seq = len(max(orig_seq_vectors, key=lambda x: len(x)))

    padded_seq_vectors = np.zeros([len(orig_seq_vectors), longest_seq, len(orig_seq_vectors[0][0])])
    padded_targets = np.zeros([len(orig_targets), longest_seq])

    for i, j in enumerate(orig_seq_vectors):
        padded_seq_vectors[i][0:len(j)] = j

    for i, j in enumerate(orig_targets):
        padded_targets[i][0:len(j)] = j

    padded_seq_vectors = torch.FloatTensor(padded_seq_vectors)
    padded_targets = torch.FloatTensor(padded_targets).view(
        (len(padded_targets), len(padded_targets[0]), 1))

    return (names, padded_seq_vectors, padded_targets)


def vector_split(v, fraction):
    """Split a vector randomly by a specified proportion

    Randomly divide the values of a vector into two, non-overlapping smaller 
    vectors. The proportions of the two vectors will be `fraction` and 
    (1 - `fraction`).

    Parameters
    ----------
    v : numpy array
            The vector to divide
    fraction : float
            Size proportion for the returned vectors. Should be in the range [0-1].

    Returns
    -------
    numpy array
            a subset of `v` of length `fraction` * len(v) (rounding up)
    numpy array
            a subset of `v` of length (1-`fraction`) * len(v). 
    """

    segment1 = np.random.choice(v, size=math.ceil(fraction*len(v)), replace=False)
    segment1.sort()
    segment2 = np.setdiff1d(v, segment1, assume_unique=True)
    return segment1, segment2


def read_split_file(split_file):
    """Read in a split_file

    Parameters
    ----------
    split_file : str
            Path to a whitespace-separated splitfile

    Returns
    -------
    numpy int array
            an array of the indices for the training set samples
    numpy int array
            an array of the indices for the validation set samples
    numpy int array
            an array of the indices for the testing set samples
    """

    with open(split_file) as f:
        lines = [line.rstrip().split() for line in f]
        training_samples = np.array([int(i) for i in lines[0]])
        val_samples = np.array([int(i) for i in lines[1]])
        test_samples = np.array([int(i) for i in lines[2]])
    return training_samples, val_samples, test_samples


def split_data(data_file, datatype, problem_type, num_classes, excludeSeqID=False, 
                split_file=None, encoding_scheme='onehot', encoder=None, 
                percent_val=0.15, percent_test=0.15, ignoreWarnings=False,
                save_splits_output=None):
    """Divide a datafile into training, validation, and test datasets

    Takes in a datafile and specification of the data format and the machine
    learning problem, and returns PyTorch-compatible Dataset objects for
    the training, validation and test sets of the data. The user may optionally
    specify how the dataset should be split into these subsets, as well as how
    protein sequences should be encoded as numeric vectors.

    Parameters
    ----------
    data_file : str
            Path to the datafile containing sequences and corresponding values
    datatype : str
            Format of the values within `data_file`. Should be 'sequence' if the
            `data_file` contains a single value per sequence, or 'residues' if
            it contains a value for each residue per sequence.
    problem_type : str
            The machine learning task to be addressed. Should be either 'regression'
            or 'classification'.
    excludeSeqID : bool, optional
            Flag that indicates how `data_file` is formatted. If False (default),
            then each line in the file should begin with a column containing a
            sequence ID. If True, then the datafile will not have this ID column,
            and will begin with the protein sequence.
    split_file : str, optional
            Path to a file containing information on how to divide the data into
            training, validation and test datasets. Default is None, which will
            cause the data to be divided randomly, with proportions based on
            `percent_val` and `percent_test`. If `split_file` is provided it must
            contain 3 lines in the file, corresponding to the training, validation
            and test sets. Each line should have whitespace-separated integer indices
            which correspond to lines in `data_file`.
    encoding_scheme : str, optional
            The method to be used for encoding protein sequences as numeric vectors.
            Currently 'onehot' and 'biophysics' are implemented (default is 'onehot').
    encoder: UserEncoder object, optional
            If encoding_scheme is 'user', encoder should be a UserEncoder object
            that can convert amino acid sequences to numeric vectors. If
            encoding_scheme is not 'user', use None.
    percent_val : float, optional
            If `split_file` is not provided, the fraction of the data that should be
            randomly assigned to the validation set. Should be in the range [0-1]
            (default is 0.15).
    percent_test : float, optional
            If `split_file` is not provided, the fraction of the data that should be
            randomly assigned to the test set. Should be in the range [0-1] (default
            is 0.15). The proportion of the training set will be calculated by the
            difference between 1 and the sum of `percent_val` and `percent_train`, so
            these should not sum to be greater than 1.
    ignoreWarnings : bool, optional
            If False, assess the structure and balance of the provided dataset with 
            basic heuristics and display warnings for common issues.
    save_splits_output : str, optional
            Location where the train / val / test splits for this run should be saved

    Returns
    -------
    SequenceDataset object
            a dataset containing the training set sequences and values
    SequenceDataset object
            a dataset containing the validation set sequences and values
    SequenceDataset object
            a dataset containing the test set sequences and values
    """

    data = parse_file(data_file, datatype, problem_type, num_classes, 
                    excludeSeqID=excludeSeqID, ignoreWarnings=ignoreWarnings)
    num_samples = len(data)

    if split_file == None:
        percent_train = 1 - percent_val - percent_test

        all_samples = np.arange(num_samples)
        training_samples, val_test_samples = vector_split(all_samples, percent_train)

        # Repeat procedure to split val and test sets
        val_test_fraction = percent_val / (percent_val + percent_test)
        val_samples, test_samples = vector_split(val_test_samples, val_test_fraction)

        # Generate datasets using these random partitions
        train_set = SequenceDataset(data=data, subset=training_samples,
                                    encoding_scheme=encoding_scheme, encoder=encoder)
        val_set = SequenceDataset(data=data, subset=val_samples,
                                  encoding_scheme=encoding_scheme, encoder=encoder)
        test_set = SequenceDataset(data=data, subset=test_samples,
                                   encoding_scheme=encoding_scheme, encoder=encoder)

        if save_splits_output != None:
            # Save train/val/test splits
            with open(save_splits_output, 'w') as out:
                out.write(" ".join(np.sort(training_samples).astype('str')))
                out.write("\n")
                out.write(" ".join(np.sort(val_samples).astype('str')))
                out.write("\n")
                out.write(" ".join(np.sort(test_samples).astype('str')))
                out.write("\n")

    else:
        training_samples, val_samples, test_samples = read_split_file(split_file)

        # Generate datasets using the provided partitions
        train_set = SequenceDataset(data=data, subset=training_samples,
                                    encoding_scheme=encoding_scheme, encoder=encoder)
        val_set = SequenceDataset(data=data, subset=val_samples,
                                  encoding_scheme=encoding_scheme, encoder=encoder)
        test_set = SequenceDataset(data=data, subset=test_samples,
                                   encoding_scheme=encoding_scheme, encoder=encoder)

    return train_set, val_set, test_set


def split_data_cv(data_file, datatype, problem_type, num_classes, excludeSeqID=False,
                  split_file=None, encoding_scheme='onehot', encoder=None,
                  percent_val=0.15, percent_test=0.15, n_folds=5, ignoreWarnings=False,
                  save_splits_output=None):
    """Divide a datafile into training, val, test and 5 cross-val datasets.

    Takes in a datafile and specification of the data format and the machine
    learning problem, and returns PyTorch-compatible Dataset objects for
    the training, validation, test and cross-validation sets of the data. The
    user may optionally specify how the dataset should be split into these 
    subsets, as well as how protein sequences should be encoded as numeric 
    vectors.

    Parameters
    ----------
    data_file : str
            Path to the datafile containing sequences and corresponding values
    datatype : str
            Format of the values within `data_file`. Should be 'sequence' if the
            `data_file` contains a single value per sequence, or 'residues' if
            it contains a value for each residue per sequence.
    problem_type : str
            The machine learning task to be addressed. Should be either 'regression'
            or 'classification'.
    excludeSeqID : bool, optional
            Flag that indicates how `data_file` is formatted. If False (default),
            then each line in the file should begin with a column containing a
            sequence ID. If True, then the datafile will not have this ID column,
            and will begin with the protein sequence.
    split_file : str, optional
            Path to a file containing information on how to divide the data into
            training, validation and test datasets. Default is None, which will
            cause the data to be divided randomly, with proportions based on
            `percent_val` and `percent_test`. If `split_file` is provided it must
            contain 3 lines in the file, corresponding to the training, validation
            and test sets. Each line should have whitespace-separated integer indices
            which correspond to lines in `data_file`.
    encoding_scheme : str, optional
            The method to be used for encoding protein sequences as numeric vectors.
            Currently 'onehot' and 'biophysics' are implemented (default is 'onehot').
    encoder: UserEncoder object, optional
            If encoding_scheme is 'user', encoder should be a UserEncoder object
            that can convert amino acid sequences to numeric vectors. If
            encoding_scheme is not 'user', use None.
    percent_val : float, optional
            If `split_file` is not provided, the fraction of the data that should be
            randomly assigned to the validation set. Should be in the range [0-1]
            (default is 0.15).
    percent_test : float, optional
            If `split_file` is not provided, the fraction of the data that should be
            randomly assigned to the test set. Should be in the range [0-1] (default
            is 0.15). The proportion of the training set will be calculated by the
            difference between 1 and the sum of `percent_val` and `percent_train`, so
            these should not sum to be greater than 1.
    n_folds : int, optional
            Number of folds for cross-validation (default is 5).
    ignoreWarnings : bool, optional
            If False, assess the structure and balance of the provided dataset with 
            basic heuristics and display warnings for common issues.
    save_splits_output : str, optional
            Location where the train / val / test splits for this run should be saved

    Returns
    -------
    list of tuples of SequenceDataset objects
            a list of tuples of length `n_folds`. Each tuple contains the training
            and validation datasets for one of the cross-val folds.
    SequenceDataset object
            a dataset containing the training set sequences and values
    SequenceDataset object
            a dataset containing the validation set sequences and values
    SequenceDataset object
            a dataset containing the test set sequences and values
    """

    data = parse_file(data_file, datatype, problem_type, num_classes, 
                    excludeSeqID=excludeSeqID, ignoreWarnings=ignoreWarnings)
    n_samples = len(data)

    # Initial step: split into training, val, and test sets
    if split_file == None:
        percent_train = 1 - percent_val - percent_test

        all_samples = np.arange(n_samples)
        training_samples, val_test_samples = vector_split(all_samples, percent_train)

        # Repeat procedure to split val and test sets
        val_test_fraction = percent_val / (percent_val + percent_test)
        val_samples, test_samples = vector_split(val_test_samples, val_test_fraction)

        # Generate datasets using these random partitions
        train_set = SequenceDataset(data=data, subset=training_samples,
                                    encoding_scheme=encoding_scheme, encoder=encoder)
        val_set = SequenceDataset(data=data, subset=val_samples,
                                  encoding_scheme=encoding_scheme, encoder=encoder)
        test_set = SequenceDataset(data=data, subset=test_samples,
                                   encoding_scheme=encoding_scheme, encoder=encoder)

        if save_splits_output != None:
            # Save train/val/test splits
            with open(save_splits_output, 'w') as out:
                out.write(" ".join(np.sort(training_samples).astype('str')))
                out.write("\n")
                out.write(" ".join(np.sort(val_samples).astype('str')))
                out.write("\n")
                out.write(" ".join(np.sort(test_samples).astype('str')))
                out.write("\n")

    # If provided, split datasets according to split_file
    else:
        training_samples, val_samples, test_samples = read_split_file(split_file)

        # Generate datasets using the provided partitions
        train_set = SequenceDataset(data=data, subset=training_samples,
                                    encoding_scheme=encoding_scheme, encoder=encoder)
        val_set = SequenceDataset(data=data, subset=val_samples,
                                  encoding_scheme=encoding_scheme, encoder=encoder)
        test_set = SequenceDataset(data=data, subset=test_samples,
                                   encoding_scheme=encoding_scheme, encoder=encoder)

    # Second step: combine train and val samples, and split evenly into n_folds
    cv_samples = np.append(training_samples, val_samples)
    np.random.shuffle(cv_samples)  # Shuffle train and val to avoid bias

    # Split into n_folds
    cv_samples = np.array_split(cv_samples, n_folds)

    # cv_sets will be a list of tuples: (fold_k_train_dataset, fold_k_test_dataset)
    cv_sets = []
    for i in range(len(cv_samples)):
        cv_test = cv_samples[i]
        cv_train = np.array([], dtype=int)
        for j in range(len(cv_samples)):
            if j != i:
                cv_train = np.append(cv_train, cv_samples[j])
        cv_train.sort()
        cv_test.sort()

        # Tuple of cross val train and test sets
        cv_sets.append((SequenceDataset(data=data, subset=cv_train,
                                        encoding_scheme=encoding_scheme, encoder=encoder),
                        SequenceDataset(data=data, subset=cv_test,
                                        encoding_scheme=encoding_scheme, encoder=encoder)))

    return cv_sets, train_set, val_set, test_set



import pytest
import parrot
from parrot import process_input_data
from parrot.parrot_exceptions import IOExceptionParrot

datapath = parrot.get_directory()
test_datapath = parrot.get_directory('../tests/test_data')



# ..........................................................................................
#
#
def test_parrot_read_sequence():
    lines = process_input_data.read_tsv_raw(f'{datapath}/res_class_dataset.tsv') 

    # expect this file to have 300 lines in it
    assert len(lines) == 300

    # check we 
    lines2 = process_input_data.read_tsv_raw(f'{datapath}/res_class_dataset.tsv', delimiter=' ' ) 

    # expect this file to have 300 lines in it
    assert len(lines2) == 300



# ..........................................................................................
#
#
def test_parrot_parse_lines():
    lines = process_input_data.read_tsv_raw(f'{datapath}/res_class_dataset.tsv') 

    # test we can convert this to 
    data = process_input_data.__parse_lines(lines, datatype='residues')



# ..........................................................................................
#
#
def test_parrot_parse_lines_error():
    lines = process_input_data.read_tsv_raw(f'{test_datapath}/res_class_dataset_broken_1.tsv') 

    # check that this bad file raises an appropriate exception
    with pytest.raises(IOExceptionParrot):
        data = process_input_data.__parse_lines(lines, datatype='residues')

    lines = process_input_data.read_tsv_raw(f'{test_datapath}/res_class_dataset_broken_2.tsv') 

    # check that this bad file raises an appropriate exception
    with pytest.raises(IOExceptionParrot):
        data = process_input_data.__parse_lines(lines, datatype='residues')




"""
Unit and regression tests for the PARROT package. Note that these tests are by no means comprehensive.
For unexplained issues, please reach out on the GitHub page:
https://github.com/idptools/parrot/issues
"""

# Import package, test suite, and other packages as needed
import parrot
import pytest
import sys
import os
import pathlib
import shutil


def test_parrot_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "parrot" in sys.modules


def test_noarg_train_call():
    """Calling parrot-train with no arguments should return an error"""
    os.chdir(pathlib.Path(__file__).parent.absolute())
    loc = shutil.which("parrot-train")
    script_descriptor = open(os.path.abspath(loc))
    script = script_descriptor.read()
    sys.argv = ["parrot-train"]

    with pytest.raises(SystemExit):
        exec(script)

    script_descriptor.close()


def test_train_invalid_dtype():
    """Calling parrot-train with invalid datatype arg should return an error"""
    os.chdir(pathlib.Path(__file__).parent.absolute())
    loc = shutil.which("parrot-train")
    script_descriptor = open(os.path.abspath(loc))
    script = script_descriptor.read()
    sys.argv = ["parrot-train", "../data/seq_class_dataset.tsv",
                "../data/output_network.pt", "-d", "gibberish", "-c", "4"]

    with pytest.raises(ValueError):
        exec(script)

    script_descriptor.close()


def test_train_invalid_numclasses():
    """Calling parrot-train with invalid number of classes should return an error"""
    os.chdir(pathlib.Path(__file__).parent.absolute())
    loc = shutil.which("parrot-train")
    script_descriptor = open(os.path.abspath(loc))
    script = script_descriptor.read()
    sys.argv = ["parrot-train", "../data/seq_class_dataset.tsv",
                "../data/output_network.pt", "-d", "sequence", "-c", "2"]

    with pytest.raises(ValueError):
        exec(script)

    script_descriptor.close()


def test_train_invalid_setFractions():
    """Calling parrot-train with --setFractions not summing to 1 should return an error"""
    os.chdir(pathlib.Path(__file__).parent.absolute())
    loc = shutil.which("parrot-train")
    script_descriptor = open(os.path.abspath(loc))
    script = script_descriptor.read()
    sys.argv = ["parrot-train", "../data/seq_class_dataset.tsv", "../data/output_network.pt",
                "-d", "sequence", "-c", "3", "--set-fractions", "0.8", "0.15", "0.15"]

    with pytest.raises(ValueError):
        exec(script)

    script_descriptor.close()


def test_noarg_optimize_call():
    """Calling parrot-optimize with no arguments should return an error"""
    os.chdir(pathlib.Path(__file__).parent.absolute())
    loc = shutil.which("parrot-optimize")
    script_descriptor = open(os.path.abspath(loc))
    script = script_descriptor.read()
    sys.argv = ["parrot-optimize"]

    with pytest.raises(SystemExit):
        exec(script)

    script_descriptor.close()


def test_optimize_invalid_dtype():
    """Calling parrot-optimize with invalid datatype arg should return an error"""
    os.chdir(pathlib.Path(__file__).parent.absolute())
    loc = shutil.which("parrot-optimize")
    script_descriptor = open(os.path.abspath(loc))
    script = script_descriptor.read()
    sys.argv = ["parrot-optimize", "../data/seq_class_dataset.tsv",
                "../data/output_network.pt", "-d", "gibberish", "-c", "4"]

    with pytest.raises(ValueError):
        exec(script)

    script_descriptor.close()


def test_optimize_invalid_numclasses():
    """Calling parrot-optimize with invalid number of classes should return an error"""
    os.chdir(pathlib.Path(__file__).parent.absolute())
    loc = shutil.which("parrot-optimize")
    script_descriptor = open(os.path.abspath(loc))
    script = script_descriptor.read()
    sys.argv = ["parrot-optimize", "../data/seq_class_dataset.tsv",
                "../data/output_network.pt", "-d", "sequence", "-c", "2"]

    with pytest.raises(ValueError):
        exec(script)

    script_descriptor.close()


def test_split_data():
    """Test that split_data() function is working as intended"""
    from parrot import process_input_data as pid

    data_file = os.path.abspath("../data/seq_class_dataset.tsv")
    train, val, test = pid.split_data(data_file, datatype='sequence',
                                  problem_type='classification', num_classes=3)

    assert (len(train) == 210) and (len(val) == 45) and (len(test) == 45) and (len(train[0]) == 3)


def test_split_data_cv():
    """Test that split_data_cv() function is working as intended"""
    from parrot import process_input_data as pid

    data_file = os.path.abspath("../data/seq_class_dataset.tsv")
    cvs, train, val, test = pid.split_data_cv(data_file, datatype='sequence',
                                  problem_type='classification', num_classes=3)

    assert (len(train) == 210) and (len(val) == 45) and (len(test) == 45) and (
        len(train[0]) == 3) and (len(cvs) == 5) and (len(cvs[0]) == 2)



"""
This file contains code for conducting Bayesian optimization.

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import math

import numpy as np

try:
    import GPy
    import GPyOpt
    from GPyOpt.methods import BayesianOptimization
except ImportError:
    print('Error importing GPy.')
    print(' If trying to run parrot-optimize, make sure to use `pip install idptools-parrot[optimize]`')

from parrot import train_network
from parrot import brnn_architecture


class BayesianOptimizer(object):
    """A class for conducting Bayesian Optimization on a PyTorch RNN

    Sets up and runs GPy Bayesian Optimization in order to choose the best-
    performing hyperparameters for a RNN for a given machine learning task. 
    Iteratively change learning rate, hidden vector size, and the number of layers
    in the network, then train and validating using 5-fold cross validation.

    Attributes
    ----------
    cv_dataloaders : list of tuples of PyTorch DataLoader objects
            For each of the cross-val folds, a tuple containing a training set
            DataLoader and a validation set DataLoader.
    input_size : int
            Length of the amino acid encoding vectors
    n_epochs : int
            Number of epochs to train for each iteration of the algorithm
    n_classes : int
            Number of classes
    n_folds : int
            Number of cross-validation folds
    problem_type : str
            'classification' or 'regression'
    dtype : str
            'sequence' or 'residues'
    weights_file : str
            Path to which the network weights will be saved during training
    device : str
            'cpu' or 'cuda' depending on system hardware
    max_iterations : int
            Maximum number of iterations to perform the optimization procedure
    silent : bool
            If true, do not print updates to console
    bds : list of dicts
            GPy-compatible bounds for each of the hyperparameters to be optimized
    """

    def __init__(self, cv_dataloaders, input_size, n_epochs, n_classes,
                 dtype, weights_file, max_iterations, device, silent):
        """
        Parameters
        ----------
        cv_dataloaders : list of tuples of PyTorch DataLoader objects
                For each of the cross-val folds, a tuple containing a training set
                DataLoader and a validation set DataLoader.
        input_size : int
                Length of the amino acid encoding vectors
        n_epochs : int
                Number of epochs to train for each iteration of the algorithm
        n_classes : int
                Number of classes
        dtype : str
                'sequence' or 'residues'
        weights_file : str
                Path to which the network weights will be saved during training
        device : str
                'cpu' or 'cuda' depending on system hardware
        max_iterations : int
                Maximum number of iterations to perform the optimization procedure
        silent : bool
                If true, do not print updates to console
        """

        self.cv_loaders = cv_dataloaders
        self.input_size = input_size
        self.n_epochs = n_epochs
        self.n_folds = len(cv_dataloaders)
        self.n_classes = n_classes
        if n_classes > 1:
            self.problem_type = 'classification'
        else:
            self.problem_type = 'regression'

        self.dtype = dtype
        self.weights_file = weights_file
        self.max_iterations = max_iterations
        self.device = device
        self.silent = silent

        self.bds = [{'name': 'log_learning_rate', 'type': 'continuous', 'domain': (-5, -2)},  # 0.00001-0.01
                    {'name': 'n_layers', 'type': 'discrete', 'domain': tuple(range(1, 6))},  # 1-5
                    {'name': 'hidden_size', 'type': 'discrete', 'domain': tuple(range(5, 51))}]  # 5-50

    def compute_cv_loss(self, hyperparameters):
        """Compute the average cross-val loss for a given set of hyperparameters

        Given N sets of hyperparameters, determine the average cross-validation loss
        for BRNNs trained with these parameters.

        Parameters
        ----------
        hyperparameters : numpy float array
                Each row corresponds to a set of hyperparameters, in the order:
                [log_learining_rate, n_layers, hidden_size]

        Returns
        -------
        numpy float array
                a Nx1 numpy array of the average cross-val loss 
                per set of input hyperparameters
        """

        cv_outputs = np.zeros((len(hyperparameters), self.n_folds))

        for i in range(len(hyperparameters)):

            log_lr, nl, hs = hyperparameters[i]
            lr = 10**float(log_lr)
            nl = int(nl)
            hs = int(hs)

            # Train and validate network with these hyperparams using k-fold CV
            cv_outputs[i] = self.eval_cv_brnns(lr, nl, hs)
            avg = np.average(cv_outputs[i])

            if self.silent is False:
                print('  %.6f	|     %2d       |         %2d           |    %.3f' % (lr, nl, hs, avg))

        outputs = np.average(cv_outputs, axis=1)
        return outputs

    def eval_cv_brnns(self, lr, nl, hs):
        """Train and test a network with given parameters across all cross-val folds

        Parameters
        ----------
        lr : float
                Learning rate of the network
        nl : int
                Number of hidden layers (for each direction) in the network
        hs : int
                Size of hidden vectors in the network

        Returns
        -------
        numpy float array
                the best validation loss from each fold of cross validation
        """

        cv_losses = np.zeros(self.n_folds) - 1  # -1 so that it's obvious if something goes wrong

        for k in range(self.n_folds):
            if self.dtype == 'sequence':
                # Use a many-to-one architecture
                brnn_network = brnn_architecture.BRNN_MtO(self.input_size, hs, nl,
                                                          self.n_classes, self.device).to(self.device)
            else:
                # Use a many-to-many architecture
                brnn_network = brnn_architecture.BRNN_MtM(self.input_size, hs, nl,
                                                          self.n_classes, self.device).to(self.device)

            # Train network with this set of hyperparameters
            train_losses, val_losses = train_network.train(brnn_network, self.cv_loaders[k][0],
                                                           self.cv_loaders[k][1], self.dtype, self.problem_type,
                                                           self.weights_file, stop_condition='iter', device=self.device,
                                                           learn_rate=lr, n_epochs=self.n_epochs, silent=True)
            # Take best val loss
            best_val_loss = np.min(val_losses)
            cv_losses[k] = best_val_loss

        return cv_losses

    def initial_search(self, x):
        """Calculate loss and estimate noise for an initial set of hyperparameters

        Parameters
        ----------
        x : numpy array
                Array containing initial hyperparameters to test

        Returns
        -------
        numpy array
                Array containing the average losses of the input hyperparameters
        float
                The standard deviation of loss across cross-val folds for the
                input hyperparameters; an estimation of the training noise
        """
        cv_outputs = np.zeros((len(x), self.n_folds))
        y = np.zeros((len(x), 1))

        for i in range(len(x)):

            log_lr, nl, hs = x[i]
            lr = 10**float(log_lr)
            nl = int(nl)
            hs = int(hs)

            # Train and validate network with these hyperparams using k-fold CV
            cv_outputs[i] = self.eval_cv_brnns(lr, nl, hs)

        # Calculate the avg and standard deviation of the losses
        for i in range(len(cv_outputs)):
            y[i] = np.average(cv_outputs[i])
        stddevs = np.std(cv_outputs, axis=1)
        avg_stddev = np.average(stddevs)

        return y, avg_stddev

    def optimize(self):
        """Set up and run Bayesian Optimization on the BRNN using GPy

        Returns
        -------
        list
                the best hyperparameters are chosen by Bayesian Optimization. Returned
                in the order: [lr, nl, hs]
        """

        # Initial hyperparameter search -- used to get noise estimate
        x_init = np.array([[-3.0, 1, 20], [-3.0, 2, 20], [-3.0, 3, 20], [-3.0, 4, 20], [-3.0, 5, 20],
                           [-2.0, 2, 20], [-3.3, 2, 20], [-4.0, 2, 20], [-5.0, 2, 20],
                           [-3.0, 2, 5],  [-3.0, 2, 15], [-3.0, 2, 35], [-3.0, 2, 50]])
        y_init, noise = self.initial_search(x_init)

        if self.silent is False:
            print("\nInitial search results:")
            print("lr\tnl\ths\toutput")
            for i in range(len(x_init)):
                print("%.5f\t%2d\t%2d\t%.4f" % (10**x_init[i][0], x_init[i][1], x_init[i][2], y_init[i][0]))
            print("Noise estimate:", noise)
            print('\n')
            print('Primary optimization:')
            print('--------------------\n')
            print('Learning rate   |   n_layers   |   hidden vector size |  avg CV loss  ')
            print('======================================================================')

        optimizer = BayesianOptimization(f=self.compute_cv_loss,
                                         domain=self.bds,
                                         model_type='GP',
                                         acquisition_type='EI',
                                         acquisition_jitter=0.05,
                                         X=x_init,
                                         Y=y_init,
                                         noise_var=noise,
                                         maximize=False)

        optimizer.run_optimization(max_iter=self.max_iterations)

        ins = optimizer.get_evaluations()[0]
        outs = optimizer.get_evaluations()[1].flatten()

        if self.silent is False:
            print("\nThe optimal hyperparameters are:\nlr = %.5f\nnl = %d\nhs = %d" %
                  (10**optimizer.x_opt[0], optimizer.x_opt[1], optimizer.x_opt[2]))
            print()

        return optimizer.x_opt



"""
Module with functions for checking if dataset meets common heuristics. Writes
warnings to console.

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import numpy as np
from scipy.stats import skew

# Duplicate sequences:
def check_duplicate_sequences(d):
    """
    Function that checks for duplicate sequences in the dataset.

    Parameters
    ------------
    d : list
        d is the parsed data from parrot.process_input_data.parse_file(). It is a 

    Returns
    -----------
    None
        No return but prints a warning if duplicate sequences are found


    """
    seen = set()
    nonuniq = set()

    # cycle over every sequence in the file
    for x in d:

        # x[1] is the actual amino acid sequence
        if x[1] not in seen:
            seen.add(x[1])
        else:
            nonuniq.add(x[1])

    nonuniq = list(nonuniq)
    if len(nonuniq) > 0:
        print("\n#############################################")
        print("WARNING: Duplicate sequences detected in dataset.\n")
        print("Duplicated sequences:")
        for seq in nonuniq:
            print(f'Duplicate found: {seq}')
        print()
        print("This may cause overfitting or overestimation of performance.\n")
        print("#############################################\n")



# Imbalanced classification dataset:
def check_class_imbalance(d):
    all_labels = np.hstack([x[2] for x in d])
    label_counts = {}
    for label in all_labels:
        if label in label_counts:
            label_counts[label] += 1
        else:
            label_counts[label] = 1

    # Check for class imbalance:
    # Balanced = 1:1 - 1:1.4
    # Slight imbalance = 1:1.4 - 1:2.5
    # Severe imbalance = 1:2.5+
    # (These divisions are arbitrary)
    min_count = min(list(label_counts.values()))
    max_count = max(list(label_counts.values()))
    total_counts = np.sum(list(label_counts.values()))
    ratio = max_count / min_count

    # Slight imbalance
    if 1.4 < ratio <= 2.5:
        print("\n#############################################")
        print("WARNING: Slight class imbalance detected in dataset.\n")
        print("Class frequencies:")
        for label, count in label_counts.items():
            print(f'  {int(label)} : {np.around(100*count/total_counts, 1)}%')
        print()
        print("This may skew predictions towards overrepresented classes.\n")
        print("#############################################\n")

    # Severe imbalance
    elif 2.5 < ratio:
        print("\n#############################################")
        print("WARNING: Severe class imbalance detected in dataset.\n")
        print("Class frequencies:")
        for label, count in label_counts.items():
            print(f'  {int(label)} : {np.around(100*count/total_counts, 1)}%')
        print()
        print("Predictions will be strongly skewed towards overrepresented classes.")
        print("Classification not recommended with current dataset.\n")
        print("#############################################\n")

# Distribution of regression values
def check_regression_imbalance(d):
    vals = np.hstack([x[2] for x in d])
    check_zero_centered(vals)
    check_low_std(vals)
    check_skew(vals)

# Regression mean location
def check_zero_centered(v):
    # Using [-5, 5] is arbitrary, but seems to do the trick
    if np.mean(v) <= -5 or np.mean(v) >= 5:
        print("\n#############################################")
        print("WARNING: Dataset not centered around zero.\n")
        print(f"Data mean: {np.around(v.mean(), 2)}")
        print()
        print("This may decrease training efficiency.\n")
        print("#############################################\n")

# Regression standard deviation magnitude
def check_low_std(v):
    # Using >5 is arbitrary, but seems to do the trick
    if np.std(v) > 5:
        print("\n#############################################")
        print("WARNING: Dataset has large variance\n")
        print(f"Data standard deviation: {np.around(np.std(v), 2)}")
        print()
        print("This may decrease training efficiency.")
        print("Consider standardizing your dataset.\n")
        print("#############################################\n")       

# Regression distribution of values
def check_skew(v):
    s = skew(v)
    if s > 1:
        print("\n#############################################")
        print("WARNING: Dataset is skewed.\n")
        print()
        print("This may cause model bias.\n")
        print("#############################################\n")

# Batch size vs dataset size:
def eval_batch_size(batch, n_samples):
    if batch > 512:
        print("\n#############################################")
        print("WARNING: Using a very large batch size.\n")
        print()
        print("This may decrease model accuracy.\n")
        print("#############################################\n")

    if n_samples // batch <= 10:
        print("\n#############################################")
        print("WARNING: Batch size is large relative to the number of samples in the training set.\n")
        print()
        print("This may decrease training efficiency.\n")
        print("#############################################\n")



"""
File containing functions for encoding a string of amino acids into a numeric vector.

.............................................................................
parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import sys
import os

import numpy as np
import torch

ONE_HOT = { 'A': 0, 'C': 1, 'D': 2, 'E': 3, 'F': 4, 'G': 5, 'H': 6, 
            'I': 7, 'K': 8, 'L': 9, 'M': 10, 'N': 11, 'P': 12, 'Q': 13, 
            'R': 14, 'S': 15, 'T': 16, 'V': 17, 'W': 18, 'Y': 19 }

def one_hot(seq):
    """Convert an amino acid sequence to a PyTorch tensor of one-hot vectors

    Each amino acid is represented by a length 20 vector with a single 1 and
    19 0's Inputing a sequence with a nono-canonical amino acid letter will
    cause the program to exit.

    E.g. Glutamic acid (E) is encoded: [0 0 0 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0]

    Parameters
    ----------
    seq : str
            An uppercase sequence of amino acids (single letter code)

    Returns
    -------
    torch.IntTensor
            a PyTorch tensor representing the encoded sequence
    """

    l = len(seq)
    m = np.zeros((l, 20))
    try:
        for i in range(l):
            m[i, ONE_HOT[seq[i]]] = 1
    except:
        error_str = 'Invalid amino acid detected: ' + seq[i]
        raise ValueError(error_str)
    return torch.from_numpy(m)


def rev_one_hot(seq_vectors):
    """Decode a list of one-hot sequence vectors into amino acid sequences

    Parameters
    ----------
    seq_vectors : list of numpy arrays
            A list containing sequence vectors

    Returns
    -------
    list
            Strings of amino acid sequences
    """

    REV_ONE_HOT = 'ACDEFGHIKLMNPQRSTVWY'
    sequences = []

    for seq_vector in seq_vectors:
        seq = []
        for residue in seq_vector:
            seq.append(REV_ONE_HOT[np.argmax(residue)])
        sequences.append("".join(seq))

    return sequences


############  Biophysical scale encoding  #############

# Map each amino acid to a vector of biophysical properties
# 0: Hydrophobicity
# 1: Charge
# 2: pI
# 3: Molecular weight (g/mol)
# 4: Aromatic amino acid
# 5: Capable of hydrogen bonding
# 6: Side chain SASA (measured from ACE-XX-NME dipeptide)
# 7: Backbone SASA (measured from ACE-XX-NME dipeptide)
# 8: Free energy of solvation
BIOPHYSICS = {  'A': [1.8,  0,  6.0,  89.1, 0, 0,  75.8,  76.1,    1.9],
                'C': [2.5,  0,  5.1, 121.2, 0, 0, 115.4,  67.9,   -1.2],
                'D': [-3.5, -1,  2.8, 133.1, 0, 1, 130.3,  71.8, -107.3],
                'E': [-3.5, -1,  3.2, 147.1, 0, 1, 161.8,  68.1, -107.3],
                'F': [2.8,  0,  5.5, 165.2, 1, 0, 209.4,  66.0,   -0.8],
                'G': [-0.4,  0,  6.0,  75.1, 0, 0,   0.0, 115.0,    0.0],
                'H': [-3.2,  1,  7.6, 155.2, 0, 1, 180.8,  67.5,  -52.7],  # Avg of HIP and HIE
                'I': [4.5,  0,  6.0, 131.2, 0, 0, 172.7,  60.3,    2.2],
                'K': [-3.9,  1,  9.7, 146.2, 0, 1, 205.9,  68.7, -100.9],
                'L': [3.8,  0,  6.0, 131.2, 0, 0, 172.0,  64.5,    2.3],
                'M': [1.9,  0,  5.7, 149.2, 0, 0, 184.8,  67.8,   -1.4],
                'N': [-3.5,  0,  5.4, 132.1, 0, 1, 142.7,  66.8,   -9.7],
                'P': [-1.6,  0,  6.3, 115.1, 0, 0, 134.3,  55.8,    2.0],
                'Q': [-3.5,  0,  5.7, 146.2, 0, 1, 173.3,  66.6,   -9.4],
                'R': [-4.5,  1, 10.8, 174.2, 0, 1, 236.5,  66.7, -100.9],
                'S': [-0.8,  0,  5.7, 105.1, 0, 1,  95.9,  72.9,   -5.1],
                'T': [-0.7,  0,  5.6, 119.1, 0, 1, 130.9,  64.1,   -5.0],
                'V': [4.2,  0,  6.0, 117.1, 0, 0, 143.1,  61.7,    2.0],
                'W': [-0.9,  0,  5.9, 204.2, 1, 1, 254.6,  64.3,   -5.9],
                'Y': [-1.3,  0,  5.7, 181.2, 1, 1, 222.5,  71.9,   -6.1]
                }

def biophysics(seq):
    """Convert an amino acid sequence to a PyTorch tensor with biophysical encoding

    Each amino acid is represented by a length 9 vector with each value representing
    a biophysical property. The nine encoded biophysical scales are Kyte-Doolittle
    hydrophobicity, charge, isoelectric point, molecular weight, aromaticity, 
    h-bonding ability, side chain solvent accessible surface area, backbone SASA, and 
    free energy of solvation. Inputing a sequence with a nono-canonical amino acid 
    letter will cause the program to exit.

    E.g. Glutamic acid (E) is: [-3.5, -1,  3.2, 147.1, 0, 1, 161.8,  68.1, -107.3]

    Parameters
    ----------
    seq : str
            An uppercase sequence of amino acids (single letter code)

    Returns
    -------
    torch.FloatTensor
            a PyTorch tensor representing the encoded sequence
    """
    l = len(seq)
    m = np.zeros((l, len(BIOPHYSICS['A'])))
    try:
        for i in range(l):
            m[i] = BIOPHYSICS[seq[i]]
    except:
        error_str = 'Invalid amino acid detected: ' + seq[i]
        raise ValueError(error_str)
    return torch.from_numpy(m)


def rev_biophysics(seq_vectors):
    """Decode a list of biophysically-encoded sequence vectors into amino acid sequences

    Parameters
    ----------
    seq_vectors : list of numpy arrays
            A list containing sequence vectors

    Returns
    -------
    list
            Strings of amino acid sequences
    """

    REV_BIOPHYSICS = {}
    for key, value in BIOPHYSICS.items():
        REV_BIOPHYSICS[str(value[6])] = key

    sequences = []
    for seq_vector in seq_vectors:
        seq = []
        for residue in seq_vector:
            seq.append(REV_BIOPHYSICS[str(residue[6])])
        sequences.append("".join(seq))

    return sequences


################## User-specified encoding ####################

def parse_encode_file(file):
    """Helper function to convert an encoding file into key:value dictionary"""

    with open(file) as f:
        lines = [x.strip().split() for x in f]

    l = len(lines[0]) - 1
    d = {}
    for line in lines:
        d[line[0]] = line[1:]

        if len(line) - 1 != l:
            raise ValueError('Vectors in encoding file do not have same length.')

    return d, l


class UserEncoder():
    """User-specified amino acid-to-vector encoding scheme object

    Attributes
    ----------
    encode_file : str
            A path to a file that describes the encoding scheme
    encode_dict : dict
            A dictionary that maps each amino acid to a numeric vector
    input_size : int
            The length of the encoding vector used for each amino acid
    """

    def __init__(self, encode_file):
        """
        Parameters
        ----------
        encode_file : str
                A path to a file that describes the encoding scheme
        """

        self.encode_file = os.path.abspath(encode_file)
        if not os.path.isfile(self.encode_file):
            raise FileNotFoundError('Encoding file does not exist.')

        self.encode_dict, self.input_size = parse_encode_file(self.encode_file)

    def __len__(self):
        """Get length of encoding scheme"""

        return self.input_size

    def encode(self, seq):
        """Convert an amino acid sequence into this encoding scheme

        Parameters
        ----------
        seq : str
                An uppercase sequence of amino acids (single letter code)

        Returns
        -------
        torch.FloatTensor
                a PyTorch tensor representing the encoded sequence
        """

        l = len(seq)
        m = np.zeros((l, self.input_size))

        try:
            for i in range(l):
                m[i] = self.encode_dict[seq[i]]
        except:
            error_str = 'Invalid amino acid detected: ' + seq[i]
            raise ValueError(error_str)
        return torch.from_numpy(m)

    def decode(self, seq_vectors):
        """Converts a list of sequence vectors back to a list of protein sequences

        Parameters
        ----------
        seq_vectors : list of numpy arrays
                A list containing sequence vectors

        Returns
        -------
        list
                Strings of amino acid sequences
        """

        # Create a reverse of the encode_dict using hashing
        rev_dict = {}
        for key, value in self.encode_dict.items():
            value = np.array(value, dtype=np.float32)
            rev_dict[hash(tuple(value))] = key

        sequences = []
        for seq_vector in seq_vectors:
            seq = []
            for residue in seq_vector:
                seq.append(rev_dict[hash(tuple(residue))])
            sequences.append("".join(seq))

        return sequences



"""
Module with functions used in parrot-preprocess

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""


# .................................................................
#
#
def are_lists_identical(l1,l2):
    """
    Quick function that asks if two lists are identical or not.

    Parameters
    -----------
    l1 : list
        First list

    l2 : list
        Second list

    Returns
    ----------
    bool
        Returns True if they are identical, or False if not
    
    """
    if len(l1) != len(l2):
        return False
    for i in range(len(l1)):
        if l1[i] != l2[i]:
            return False
    return True


# .................................................................
#
#
def recode_classes(lines):
    """
    Function which takes a set of data read in (lines) and converts it into a PARROT-compliant set of 
    lines files where classes start at 0. This works for both residue and sequence datatypes.

    Parameters
    -----------
    lines : list
        The lines list is a list of lists, where each sublist corresponds to a line in a parsed
        datafile. If the dataype is sequence then each sublist has three elements (id, sequence, class/score),
        while if the datatype is residues then each sublists has n elements (id, sequence, per-residue scores or 
        classes for each residue).


    Returns
    -------------
    list
        Returns a revised lines list, where sequence or residue classes are guarenteed to start at 0 and increase
        non-monnotonically.
    
    """
    
    # initialize empty vars
    classes = set([])
    class_mapping = {}


    # first figure out the number of classes
    for line in lines:
        classes = classes.union(set(line[2:]))

    n_classes = len(classes)

    if n_classes > 100:
        print('WARNING: 100 individual classes found. This is probably not a good classification problem...')

    # next build a dictionary that maps each unique class (sorted) back to a monotonically increasing
    # number of class labels starting at 0
    classes = list(classes)
    classes.sort()

    idx = 0
    for i in classes:
        class_mapping[i] = idx
        idx = idx + 1

    # Finally, build a new lines list, where the sequence or residues classes are mapped
    # via the class_mapping dictionary 
    new_lines = []
    for line in lines:

        # convert the class
        tmp = [line[0], line[1]]       
        tmp.extend([str(class_mapping[i]) for i in line[2:]])

        new_lines.append(tmp)
        
    return (new_lines, class_mapping)



# .................................................................
#
#
def remove_sequence_duplicates(lines, verbose=False):
    """
    Function that removes any line where the same sequence appears twice. Because in such
    and instance we cannot know which of the two sequences is the 'ground truth', this method
    removes both sequences from the dataset

    Parameters
    -------------
    lines : list
        Lines list file generated by parrot.process_input_data.__read_tsv(<filename>).

    verbose : bool
        Flag which tells the function how loud to be. Default = False.

    Returns
    ------------
    list
        Returns an updated lines file that has had all the lines with the same sequence
        removed - i.e. the resulting file is guarenteed to have unique sequences, although
        not necessarily unique IDs or datasets.

    """
    if verbose:
        print('-------------------------------------------------------------------')
        print('Performing duplicate sequence removal --remove-sequence-duplicates ')
        print('-------------------------------------------------------------------')


    seen = {}
    nonuniq = set([])


    duplicate_count = 0
    for line in lines:

        if line[1] not in seen:
            seen[line[1]] = line
        else:
            duplicate_count = duplicate_count + 1
            nonuniq.add(line[1])

    
    for x in nonuniq:
        seen.pop(x)

    new_lines = []
    for x in seen:
        new_lines.append(seen[x])

    if verbose:
        print(f'   Found {duplicate_count} sequence duplicates in this file. All copies of these sequences have been removed.\n')
        print(f'   NOTE this may be more aggressive than you intended - to only remove one of the duplicate copies of duplicate lines, use the --remove-line-dup flag instead')
        

        print(f'   Also found {seq_duplicate_count} ID-sequence duplicates in this file (same ID, same sequence, different data.\n   If any of these were full duplicates (likely) they will have been removed.\n')
        print(f'   Found {seq_duplicate_count_full} sequence duplicates in this file (same sequence with different ID and different data).\n   If any of these were full duplicates (likely) they will have been removed.\n')

    if verbose:
        print(f'Found {duplicate_count} sequence duplicates. These have been removed.')        

    return new_lines
                


# .................................................................
#
#
def remove_line_duplicates(lines, verbose=False):
    """
    Function that removes true duplicates and warns about potential psuedoduplicates if
    verbose is set to true.

    Parameters
    -------------
    lines : list
        Lines list file generated by parrot.process_input_data.__read_tsv(<filename>).

    verbose : bool
        Flag which tells the function how loud to be. Default = False.

    Returns
    ------------
    list
        Returns an updated lines file that has had all the true duplicate lines removed (i.e.
        lines where the id, sequence, and data were all identical. 
    

    """

    if verbose:
        print('---------------------------------------------------------------')
        print('Performing duplicate checking/removal --remove-line-duplicates ')
        print('---------------------------------------------------------------')

    found = {}
    all_sequences = set([])
    duplicate_count = 0
    seq_duplicate_count = 0
    seq_duplicate_count_full = 0

    # for each line
    for i in lines:

        # this part ONLY keeps track of sequences. Note these are not removed
        # because which one should you remove!
        if i[1] in all_sequences:
            seq_duplicate_count_full = seq_duplicate_count_full + 1
        else:
            all_sequences.add(i[1])

        # if the ID code is has not yet been seen
        if i[0] not in found:
            found[i[0]] = [i]

        # else if the ID code (first column in datafile) HAS been seen
        else:
            
            # set the unique flag to true
            this_is_a_duplicate = False

            # for EVERY time it was seen (hits is a previous line)
            for hits in found[i[0]]:

                # if the sequence ALSO MATCHED
                if hits[1] == i[1]:
                    seq_duplicate_count = seq_duplicate_count + 1
                    
                    if are_lists_identical(hits[2:], i[2:]):
                        duplicate_count = duplicate_count + 1
                        #
                        this_is_a_duplicate = True

            # if we get here then this was 
            if this_is_a_duplicate is False:
                found[i[0]].append(i)

    if verbose:
        print(f'   Found {duplicate_count} full duplicate lines in this file. These have been removed.\n')
        print(f'   Also found {seq_duplicate_count} ID-sequence duplicates in this file (same ID, same sequence, different data.\n   If any of these were full duplicates (likely) they will have been removed.\n')
        print(f'   Found {seq_duplicate_count_full} sequence duplicates in this file (same sequence with different ID and different data).\n   If any of these were full duplicates (likely) they will have been removed.\n')

    new_lines = []
    for id_key in found:
        for x in found[id_key]:
            new_lines.append(x)

    return new_lines
        
                







"""
Module with functions for dealing with automatic class balacing

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import random
import numpy as np
import datetime


from parrot.parrot_exceptions import ParrotException 

EMERGENCY_COUNT_THRESHOLD=10



def __validate_class_balance_input(balance_threshold, max_fraction_removed):
    """
    Helper function that checks that a balance_threshold and a max_fraction_removed threshold

    Parameters
    -------------
    balance_threshold : float
        Float that reports on the balance score threshold to be used. Expects to be between 1 and 5.0

    max_fraction_removed : float
        Fraction which defines the max fraction of the data that can be removed in the name of 
        class balancing. Expects to be between 0 and 1 not inclusive.


    Returns
    ----------
    None
        No return

    Raises
    ----------
    parrot.parrot_exceptions.ParrotException
        If an invalid exception then raises a ParrotException

    """
    
    if balance_threshold < 1.0 or balance_threshold > 5:
        raise ParrotException('Error during class balancing. The balance threshold must be between 1.0 and 5 - asking to balance with a balance score > 5 does not any sense')

    if max_fraction_removed <= 0 or max_fraction_removed >= 1.0:
        raise ParrotException('Error during class balancing. max_fraction_removed must be between 0 and 1 (not inclusive)')
    

# ..........................................................................................
#
#
def parse_class_info(lines):
    """
    Function that extracts out class information from a set of parsed datafile
    lines. Lines is a list of lists and should be generated by the function
    parrot.process_input_data.__read_tsv(<filename>).

    The function takes in a files list and returns a tuple with two bits of 
    information; the list of classes found in the file, and the mapping between
    sequence index and entry. Note sequence index is a numerical internal index
    that starts at 0 for the first line and increments accordingly, and does NOT
    correspond to the identifier used in the data file.

    Parameters
    ------------
    lines : list
        Lines is a specific type of list, defined as the output from the function
        parrot.process_input_data.__read_tsv(<filename>).

    Returns
    ----------
    tuple
        The function returns a tuple with the following two elements
    
        [0] - list of the classes (where classes are strings as defined in the 
              lines list)

        [1] - a dictionary that maps index (starting at 0) to a second dictionary,
              which itself maps the class name to the number of instances that
              class pops up. This dictionary is the idx2classes dict which is
              referenced through this file
    
    """
    original_data = {}
    classes = set([])
    # first figure out the number of classes
    idx=0
    for line in lines:
        classes = classes.union(set(line[2:]))
        original_data[idx] = line
        idx = idx + 1

    # order classes in a consistent manner
    n_classes = len(classes)
    classes = list(classes)
    classes.sort()
    
    idx2classes = {}
    for idx in original_data:
        tmp = {}
        for c in classes:
            tmp[c] = original_data[idx][2:].count(c)
        idx2classes[idx] = tmp

    return (classes,idx2classes)


# ..........................................................................................
#
#
def get_total_class_count(classes, idx2classes, old_total_class_count=None, removed_entry_dict=None):
    """
    Function which calculates the counts for each of the classes - i.e. given the idx2classes dictionary, how many times
    does each of the classes (defined by the classes list) appear?
    
    
    This can be calculate in one of two ways:

    (1) Directly from the idx2classes dictionary with the classes list. This is computationally a little expensive
        in that each entry in idx2classes dictionary, but gives a ground truthe.

    (2) Using an old total_class_count dictionary and a removed_entry_dict which reflects a single entry that is to
        be removed. To expand a bit on this, if an old_total_class_count dictionary is provided, it should be a dictionary
        of the format {'class name':class_count,...}. Then, a removed_entry_dict should be of the same format, an the
        returned total_class_count dictionary subtracts the counts for the removed_entry_dict from the old_total_class_count
        dictionary. 

    The first option is a de novo calculation, whereas the second one is an almost instantaneous correction, although we 
    do allocate a new dictionary so the return variable is a different instantiation than the old_total_class_count.
 
    Parameters
    -------------
    classes : list
        List of valid classes. All classes in the idx2classes dictionary should be represented in the 
        classes list, and, similarly, every dictionary in the idx2classes should map between each of 
        the classes and a count (even if the count is 0).
        
    idx2classes : dict
        Dictionary that maps index (starting at 0) to a second dictionary, which itself maps the 
        class name to the number of instances that class pops up.

    old_total_class_count : dict
        If provided, this should be a single dictionary that maps class to total count. 
        Default = None.

    removed_entry_dict : dict
        If provided, this should be a single dictionary that maps class to a count. The counts here
        are subtracted from the old_total_class_count


    Returns
    ------------
    dict
        Dict which maps between class identity and total counts for that class

    Raises
    ------------
    Exception
    
            
    """

    # if both removed entry dict and old_total_class_count are provdied
    if removed_entry_dict is not None and old_total_class_count is not None:

        # update the old_total_class_count and return - note we do allocate a 
        # new dictionary in case this function ever is used in a scenario where
        # pass by reference would be a problem        
        new_dict = {}
        for c in old_total_class_count:
            new_dict[c] = old_total_class_count[c] - removed_entry_dict[c]
            
        return new_dict
    

    # if only one of these two is allocate through an exception
    if removed_entry_dict is not None or old_total_class_count is not None:
        raise Exception('If removed_entry_dict is provided then old_total_class_count must also be provided (and vice versa)')
    

    # If we get here we're calculating the total_class_count from scratch
    total_class_count = {}

    # initialize
    for c in classes:
        total_class_count[c] = 0
        
    # for each entry
    for idx in idx2classes:
        for c in classes:
            total_class_count[c] = total_class_count[c] + idx2classes[idx][c]
                        
    return total_class_count


# ..........................................................................................
#
#
def calculate_imbalance(total_class_count, idx_to_remove=None, idx2classes=None):
    """
    Function that calculates a simple metric for class imbalance, which is just the ratio
    of the most abundant class over the least abundant class.

    A perfectly balanced dataset would have an imbalance score of 1.0. 

    If idx_to_remove and idx2classes are provided, the imbalance score is calculated
    AFTER the idx_to_remove was removed from the total_class_count
    
    Parameters
    ----------------
    total_class_count : dict
        Dict which is total count for each class. This is the output from 
        get_total_class_count.
        
    idx_to_remove : int
        If provided, defines the index of the entry that is to be removed
        prior to imbalance calculation. Default = None.
        
    idx2classes : dict
        Dict that maps an index to a second dictionary which says for that
        sequence what the count of each class is. Default = None.

    Returns
    -------------
    float
        Returns a float which is the ratio of the largest to smallest class. If idx_to_remove
        is passed then calculates what this would be after that idx was removed (but does not)
        actually remove it!

    
    """
        
    # if we're asking how removing a specific index would change the imbalance...
    if idx_to_remove is not None and idx2classes is not None:
        local_class_counts = idx2classes[idx_to_remove]
        new_total_class_count = {}
        
        for c in local_class_counts:
            new_total_class_count[c] = total_class_count[c] - local_class_counts[c]

    elif idx_to_remove is None and idx2classes is  None:
        new_total_class_count = total_class_count

    else:
        raise Exception('If idx_to_remove is passed must also pass idx2classes')


        
    biggest_idx = max(new_total_class_count, key=new_total_class_count.get)
    smallest_idx = min(new_total_class_count, key=new_total_class_count.get)

    return (new_total_class_count[biggest_idx] / new_total_class_count[smallest_idx])


        
# ..........................................................................................
#
#    
def run_class_balance_residues(lines, 
                               balance_threshold=1.4,                                
                               block_size=100, 
                               max_fraction_removed=0.25, 
                               shuffle_data=True,
                               verbose=True):

    """
    Function that takes in a read lines dataset (from parrot-preprocess) and removes
    a subset of the elements such that the dataset is class balanced. This implements the 
    Stochastic Residue Class Balancing (SRCB) algorithm, an approach to classified sequence
    data to generate a more balanced dataset.

    Specifically, SRCBworks by doing the following:

    1. Calculating an inbalance score (ratio of most abundant class over least abundant class)

    2. Asking if this value is below the balance_threshold - if yes, exit and we're done
    
    3. If no, we cycle through the data for a fixed number of iterations. Each iteration we 
       randomly select a set of block_size datapoints, and for each of those datapoints ask 
       how the imbalance score would be changed if each of them were removed. We find the datapoint
       that leads to the greates decrease in the imbalance score and then remove that data point.

    4. This process is repeated unill EITHER we removed the max number of datapoints we allow, as
       defined by max_fraction_removed, OR our new dataset has a sufficiently low imbalance score. 
       
    The big advantage of SRCB  is that it scales extremely well with data size because each
    iteration in the removal step ONLY compares a fixed (block_size) number of datapoints, 
    and changes to the imbalance score can then be calculated based on a change in the previous
    values, such that we basically avoid iteracting over the entire dataset entirely other than
    during an intial shuffle operation to remove correlations in the data and when constructing
    the initial counts for classes. This approach runs in under ~30 seconds for datasets of 
    100s of thousands of datapoints. As block_size gets bigger and max_fraction_removed gets
    larger than algoirthm will slow approximately linearly with both parameters.

    In addition to the parameters passed into this function, there's a hidden parameter defined
    as a static const in this file called EMERGENCY_COUNT_THRESHOLD. This parameter defines how 
    many times we try and find a block that improves the imbalance score - for example. If after
    EMERGENCY_COUNT_THRESHOLD attempts we stil don't find a datapoint that improves the current
    balance we'll remove a datapoint that makes it the 'least bad'. This ensures the algorithm
    cannot get stuck in an infite loop. Right now this value is set to 10, but, if it should
    be set to something more permissive I'm not sure.

    Parameters
    -------------
    lines : list
        Lines is a list of lists, where each sublist is of the format [<id>, <sequence>, <class1>, ...,<classn>]. The 
        number of class1,class2 data points should match the number of residues in the <sequence>.

    balance_threshold : float
        A value that reports on how much imbalance is tolerated. The imbalance score is just the ratio of the 
        count for the most frequenct class divided by the count of the least frequent class. i.e. a score of 1.0 would
        mean the dataset is perfectly balanced. We use a threshold of 1.4 to match the warning thresholds defined in
        parrot.tools.dataset_warnings, but these could be updated. Default = 1.4

    block_size : int
        Parameter that defines the size of the local block size. In practice, if block_size were set to the data
        size this algorithm would GUARENTEE the optimally balanced dataset. In practice, using a block size of ~0.1%
        of the data (e.g. 100 data points in a dataset of 100,000) has been reasonable in achieving an almost
        identical d(imbalance score)/d(number of data points removed) for 'real' data. Default 100.

    max_fraction_removed : float
        What fraction of the data can be removed in the persuit of a balance dataset? Must be between 0 and 1 (not
        inclusive of either). Default = 0.25.

    shuffle_data : bool
        Flag which, if set to true, means data is fully shuffled before sub-sampling is done. This is generally
        good in that the SRCB algorithm assumes no correlation in data, but there may be times in which retaining
        the approximate structure of the input data (with a subset of lines removed) is preferable, so, this
        gives some control over that.

    verbose : bool
        Flag which indicates if status and input info should be printed to screen.

    Returns
    -----------
    (list, list)
        Returns a tuple of two lists:

        [0] - The first list is the down-sampled data in the format of a standard lines list (i.e. matches the
              lines input variable).

        [1] - Trace of balance score over each iteration. Only includes iterations where a datapoint
              was removed. Useful for assessing how changing the parameters influenced the trajectory
              if data exclusion. 

    """

    # validate input...
    if block_size < 2:
        raise ParrotException('Error during class balancing. Block size must be greater than 1')

    if block_size > 0.2*len(lines):
        print(f'Warning: Block size is unecessarily large - reducing to {int(0.2*len(lines))} [10% of data size].')
        block_size = int(0.2*len(lines))

    __validate_class_balance_input(balance_threshold, max_fraction_removed)

    # shuffle lines - we do this so any intrinsic structure in the input data. The ClassBalance alrg
    if shuffle_data:
        random.shuffle(lines)

    # parse input 
    (classes, idx2classes) = parse_class_info(lines)

    # initial class imbalance score
    total_class_count = get_total_class_count(classes, idx2classes)
    original_imbalance_score = calculate_imbalance(total_class_count=total_class_count)

    itercount = int(len(idx2classes)*max_fraction_removed)

    # cross-check itercount to make sure it's a useable number
    if itercount < 1:
        itercount = 1

    if itercount > len(idx2classes):
        itercount = len(idx2classes) - 1
    

    if verbose:
        print('-------------------------------------------------------------------')
        print('Performing automatic class balancing for residues (--class-balance)') 
        print('-------------------------------------------------------------------')
        print('')
        print(f'   Found {len(classes)} different classes accross {len(idx2classes)} sequences')
        print('   Class breakdown:')
        print("   " + str(total_class_count))
        print(f'   These classes have an imbalance ratio of {original_imbalance_score}')
        print(f'   Preparing to perform class balancing with the following parameters')
        print(f'   Max num. of removable sequences : {itercount}')
        print(f'   Imbalance threshold             : {balance_threshold}')
        print(f'   Block size                      : {block_size}')
        print(f'   Starting at {datetime.datetime.now()}') 


    # initialize a bunch of stuff
    removed_entry = None
    tracker = []
    tracker.append(original_imbalance_score)
    c = 0 
    emergency_count = 0
    recalculate_total_class_count = True

    # main whileloop of SRCB
    while c < itercount:

        # the emergency count incremenets on each iteraction and helps us 
        # avoid getting stuck in a while loop.
        emergency_count = emergency_count + 1

        # get the current set of possible indices we have to chose from in the
        # dataset. Recall indices here correspond to line numbers in the input 
        # lines list.
        valid_keys = list(idx2classes.keys())

        # select a random value between 0 and the number of keys minus the block size
        random_start = random.randint(0, len(valid_keys) - (1+block_size))

        # select a random value between the random start and the end of the number of keys
        random_end = random_start + block_size

        # finally, select a selection of indices based on the random start/end 
        # positions. 
        subset = valid_keys[random_start:random_end]

        # calculate total class count and total number of residues. If this is a perturbation from a previously
        # calculated value we use the prior state rather than recomputing everything. Note we only do this if
        # recalculate_total_class_count is set to True. If nothing was removed we don't need to recalculate
        # this as it hasn't changed.
        if recalculate_total_class_count:
            if removed_entry is not None:
                total_class_count = get_total_class_count(classes, idx2classes, old_total_class_count=total_class_count, removed_entry_dict=removed_entry)
            else:
                total_class_count = get_total_class_count(classes, idx2classes)

        # if we have reached the threshold...
        if tracker[-1] < balance_threshold:
            break

        # calculate how imbalance score would change if we removed each of the datapoints defined by the list of indices
        # in the subset list
        hits = {}
        for idx in subset:
            inbalance_score = calculate_imbalance(total_class_count=total_class_count, idx_to_remove=idx, idx2classes=idx2classes)
            hits[idx] = inbalance_score

        # get the index of the datapoint from subset that leads to the imabalance score to be the smallest
        offending = min(hits, key=hits.get)

        # only remove this entry if it makes things better or we've been stuck on the same 
        
        if hits[offending] < tracker[-1] or emergency_count % EMERGENCY_COUNT_THRESHOLD == 0:
            
            if emergency_count % EMERGENCY_COUNT_THRESHOLD == 0:                
                if verbose:
                    print(f'WARNING: Got stuck trying to improve the balance score on iterations {c} [tried {emergency_count} times).\nConsider re-running with a lower fraction to remove or a larger block-size')

            # remove data point
            removed_entry = idx2classes.pop(offending)    

            # increment counter
            c = c + 1

            # reset emergency count counter and recalculate_total_class_count flag
            emergency_count = 0
            recalculate_total_class_count = True

            # update the tracker
            tracker.append(calculate_imbalance(total_class_count=total_class_count))
        else:
            recalculate_total_class_count = False



    # note this works because the indices in idx2classes correspond to the line
    # number in the original lines input list
    new_lines = []
    for idx in list(idx2classes.keys()):
        new_lines.append(lines[idx])
        
    if verbose:
        frac_removed = c/len(lines)
        print('   -------------------------------------')
        print(f'   Complete at {datetime.datetime.now()}') 
        print(f'   Removed {c} sequences ({frac_removed:.3f} of sequences)')
        print(f'   Updated balance score: {tracker[-1]}')
        print(f'   Updated class breakdown:')
        print("   "+str(total_class_count))
        print('   -------------------------------------\n')
        


    return (new_lines, tracker)
    

        

# ..........................................................................................
#
#    
# Note - most of this module is concerned wth run_class_balance_residue - the sequence-based
# class balancing is much easier and is dealt with internally here
#
#
def run_class_balance_sequences(lines, 
                                balance_threshold=1.4, 
                                max_fraction_removed=0.25, 
                                verbose=True):

    """
    Function that performs class balancing if classes are set at the sequence level. This uses a completely
    different algorithm to the SRCB (for sequences) as this is a much simpler problem. This algorithm here
    uses the following approach:


    1. Groups sequences based on their class
    2. Calculates imbalance score (count for most abundant class / count for least abundant class)
    3. Calculates the count for the most abundant class and the average count for the remaining classes
    4. Subtracts random datapoints from the most abundant class to match the average of the remainders
    5. Recalculate and check how many data points were removed
    6. Repeat or exit

    This is a much more simple and much faster algorithm the the SRCB. It still scales well with arbitrarily
    large datasets, in that in general only a few iterations are required for convergence because each iteration
    removes many datapoints.

    Parameters
    -------------
    lines : list
        Lines is a list of lists, where each sublist is of the format [<id>, <sequence>, <class>]. This in general
        should be generated by the function parrot.process_input_data.__read_tsv().

    balance_threshold : float
        A value that reports on how much imbalance is tolerated. The imbalance score is just the ratio of the 
        count for the most frequenct class divided by the count of the least frequent class. i.e. a score of 1.0 would
        mean the dataset is perfectly balanced. We use a threshold of 1.4 to match the warning thresholds defined in
        parrot.tools.dataset_warnings, but these could be updated. Default = 1.4
    

    max_fraction_removed : float
        What fraction of the data can be removed in the persuit of a balance dataset? Must be between 0 and 1 (not
        inclusive of either). Default = 0.25.

    verbose : bool
        Flag which indicates if status and input info should be printed to screen.
    
    Returns
    -----------
    list
        Down-sampled lines-list data in the format of a standard lines list (i.e. matches the
        lines input variable). 

    """

    
    # ================================================================================================
    def __build_total_class_count(c2l):
        """
        Internal function that converts a class2line dictonary to a total_class_count dictionary.
        This is defined as an internal function because its only appropriate to work with sequence
        classes, not residues classes.

        Parameters
        ------------
        c2l : dict
            The class2dict dictionary has class names as keys and values are lists of lines that
            correspond to that class.

        Returns
        ----------
        dict
            Returns a total_class_count dictionary, where keys are class names and values are number
            of times a line with that class appears.
        
        """
    
        new_total_class_count = {}
        for x in c2l:
            new_total_class_count[x] = len(c2l[x])
        
        new_total_class_count = {k: v for k, v in sorted(new_total_class_count.items(), key=lambda item: item[1], reverse=True)}
        return new_total_class_count
    # ================================================================================================

    
    # validate input options (same validation as used for run_class_balance_residues())
    __validate_class_balance_input(balance_threshold, max_fraction_removed)



    # shuffle lines
    random.shuffle(lines)
    

    # maps class name => list of lines that have datapoints for that class
    class2line = {}

    # maps class name => count 
    total_class_count = {} 
    
    # build the class2line dictionary and total_class_count dictionary
    for line in lines:
        c = line[2]

        if c not in class2line:
            class2line[c] = []
            total_class_count[c] = 0

        class2line[c].append(line)
        total_class_count[c] = total_class_count[c] + 1

    # this generates an ordered dictionary where the keys are largest-to-smallest ordered by count
    total_class_count = {k: v for k, v in sorted(total_class_count.items(), key=lambda item: item[1], reverse=True)}

    # calculate the max number of datapoints we're allowed to remove
    max_removal = int(len(lines)*max_fraction_removed)
        
    # calculate initial class balance score (count of biggest class/count of smallest class)
    class_balance_score = total_class_count[list(total_class_count.keys())[0]]/total_class_count[list(total_class_count.keys())[-1]]

    # set the all-time-removal counter
    all_time_removal = 0

    if verbose:
        print('-------------------------------------------------------------------')
        print('Performing automatic class balancing for sequences (--class-balance)') 
        print('-------------------------------------------------------------------')
        print('')
        print(f'   Found {len(total_class_count)} different classes accross {len(lines)} sequences')
        print('   Class breakdown:')
        print("   "+str(total_class_count))
        print(f'   These classes have an imbalance ratio of {class_balance_score}')
        print(f'   Preparing to perform class balancing with the following parameters')
        print(f'   Max num. of removable sequences: {max_removal}')
        print(f'   Imbalance threshold            : {balance_threshold}')
        print(f'   Starting at {datetime.datetime.now()}\n') 

    # main loop
    while class_balance_score > balance_threshold:

        # calculate current dataset size (this will change as we remove datapoints
        dataset_size = np.sum([total_class_count[x] for x in total_class_count])

        # determine the name of the largest and smallest classt 
        largest_class = list(total_class_count.keys())[0] # class name
        smallest_class = list(total_class_count.keys())[-1] # class name
    
        # determine the names of all the classes that are not the largest class
        not_largest_class = list(total_class_count.keys())[1:] 
    
        # compute average count of all classes except the largest class
        average_of_not_largest_class = np.mean([total_class_count[x] for x in not_largest_class])
    
        # diff is the difference between the largest class and the average of the other classes. Note we use the
        # min of 1 or this value because this avoids an edge case where the diff would round to 0 if the mean
        # was the same...
        diff = np.max([1,total_class_count[largest_class] - average_of_not_largest_class])
    
        # remove EITHER the diff between largest and average OR the maximum number of 
        # datapoints we can remove in this round, whichever is smallest
        local_removal_count = int(np.min([diff,max_removal-all_time_removal]))

        # remove a set of datapoints from the largest cluster. We shuffled everything to start
        # so this statistically is just randomly removing data, even though its the first $n$
        # from this dataset
        for i in range(local_removal_count):
            class2line[largest_class].pop()
    
        # update the total_class_count (recall key order is defined by class count and goes largest to smallest)
        total_class_count = __build_total_class_count(class2line)
    
        # update the count of how many data points we've removed
        all_time_removal = all_time_removal + local_removal_count

        # recalculate the class imbalance score
        class_balance_score = total_class_count[list(total_class_count.keys())[0]]/total_class_count[list(total_class_count.keys())[-1]]
    
        # this gives us a threshold where we exist when we removed as many datapoints as we're allowed
        if local_removal_count == (max_removal - all_time_removal):
            break


    # finally construct return data and SHUFFLE the data so we don't accidentally write out a structured file which
    # would be problematic down the line
    new_lines = []
    for c in class2line:
        new_lines.extend(class2line[c])

    random.shuffle(new_lines)

    if verbose:
        frac_removed = all_time_removal/len(lines)
        print('   -------------------------------------')
        print(f'   Complete at {datetime.datetime.now()}') 
        print(f'   Removed {all_time_removal} sequences ({frac_removed:.3f} of sequences)')
        print(f'   Updated balance score: {class_balance_score}')
        print(f'   Updated class breakdown:')
        print("   "+str(total_class_count))
        print('   -------------------------------------\n')

    return new_lines
        



import parrot

def print_startup(program_name, silent, print_parrot=False):
    """
    Function that prints a welcoming startup message with a 
    snazzy picture of a parrot.
    
    Parameters
    -----------
    program_name : str
        Name of the cli program being executed

    print_parrot : bool
        If set to True, a snazzy parrot pictures is printed

    silent : bool
        Boolean which, if set to true, means no message is printed 

    Returns
    --------
    None
        No return but will print a startup message unless silent=True

    
    """

    # if silent print nothing
    if silent:
        return

    if print_parrot is True:
        print("...................................")
        print("|                                 |")
        print("|    _        Welcome to PARROT!  |")
        print("|  /` '\                          |")    
        print("|/| @   l                         |")
        print("|\|      \                        |")
        print("|  `\     `\_                     |")
        print("|    \    __ `\                   |")
        print("|    l  \   `\ `\__               |")
        print("|     \  `\./`     ``\            |")
        print("|       \ ____ / \   l            |")
        print("|         ||  ||  )  /            |")
        print("|-------(((-(((---l /-------------|")
        print("|                l /              |")
        print("|               / /               |")
        print("|              / /                |")
        print("|             //                  |")
        print("|            /                    |")
        print("...................................")

    print(f'\nparrot-predict (version {parrot.__version__})')
    print("...................................")
    print(f'Developed by Dan Griffith\n   Holehouse lab, WUSTL')
    print("")
    print("")
    


def print_settings(silent, buffer_length=14, **kwargs):
    """
    Function that dynamically prints out a list of key-value pairs
    as passed. Means that when this function can be 
    


    """


    # ............................................................
    def strbuffer(local_string):
        """
        Short internal function that pads a string so it is 
        buffer_length long. Note BUFFER_LENGTH is set when the
        function is defined
        """

        # 
        buffer_count = buffer_length - len(local_string)

        if buffer_count < 0:
            return local_string

        return local_string + " "*buffer_count
    # ............................................................



    if silent:
        return

    print('Settings:')
    print("...................................")
    for name in kwargs:
        name_buffered = strbuffer(name)
        print(f"{name_buffered}: {kwargs[name]}")



# .................................................................
#
#
def write_datafile(filename, outdata):
    """
    Function that writes out a parrot datafile, where outdata must
    be a list of lists and each sub-element is joined on a whitespace

    Parameters
    --------------
    filename : str
        The name of the file to be written to

    outdata : list
        A list of lists (e.g. lines in parrot-preprocess)

    Returns
    ---------
    None
        No returns but the outdata is written to the current file

    """
    with open(filename,'w') as fh:
        for line in outdata:
            fh.write(" ".join(line)+"\n")


# .................................................................
#
#
def validate_args(passed, valid, error_message=''):
    """
    General function that validates a specific string against a list of strings
    and raises an exception if the string ($passed) is not found in the list (valid).
    Good for validating passed options.

    Parameters
    -----------
    passed : str
        String of interest.

    valid : list
        List of strings

    Returns
    ----------
    None

    Raises
    -----------
    Exception
        If passed is not in valid, raise an exception
    
    """
    if passed not in valid:
        raise Exception(f'ERROR: Invalid option passed {passed}.\n{error_message}')


        



'''
Python module for integrating a trained network directly into a Python workflow.

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 

'''

from parrot import brnn_architecture
from parrot import encode_sequence

import torch
import numpy as np
import os

def softmax(v):
    return (np.e ** v) / np.sum(np.e ** v)

class Predictor():
    """Class that for integrating a trained PARROT network into a Python workflow

    Usage:
    >>> from parrot import py_predictor
    >>> my_predictor = py_predictor.Predictor(</path/to/saved_network.pt>, 
                                                dtype={"sequence" or "residues"})
    >>> value = my_predictor.predict(AA_sequence)

    ***
    NOTE:   Assumes all sequences are composed of canonical amino acids and 
            that all networks were implemented using one-hot encoding.
    ***


    Attributes
    ----------
    dtype : str
            Data format that the network was trained for. Either "sequence" or 
            "residues".
    num_layers : int
            Number of hidden layers in the trained network.
    hidden_vector_size : int
            Size of hidden vectoer in the trained network.
    n_classes : int
            Number of data classes that the network was trained for. If 1, then
            network is designed for regression task. If >1, then classification
            task with n_classes.
    task : str
            Designates if network is designed for "classification" or "regression".
    network : PyTorch object
            Initialized PARROT network with loaded weights.
    """

    def __init__(self, saved_weights, dtype):
        """
        Parameters
        ----------
        saved_weights : str or Path
                Location of the saved network weights. If a valid file is provided,
                network parameters will be dynamically read in an network will be 
                initialized.
        dtype : str
                Data format that the network was trained for. Either "sequence" or 
                "residues".
        """

        self.dtype = dtype
  
        loaded_model = torch.load(saved_weights, map_location=torch.device('cpu'))
      
        # Dynamically read in correct network size:
        self.num_layers = 0
        while True:
            s = f'lstm.weight_ih_l{self.num_layers}'
            try:
                temp = loaded_model[s]
                self.num_layers += 1
            except KeyError:
                break
        
        # Extract other network hyperparams           
        self.hidden_vector_size = int(np.shape(loaded_model['lstm.weight_ih_l0'])[0] / 4)
        self.n_classes = np.shape(loaded_model['fc.bias'])[0]

        if self.n_classes > 1:
            self.task = "classification"
        else:
            self.task = "regression"

        # Instantiate network weights into Predictor() object
        if self.dtype == "sequence":
            self.network = brnn_architecture.BRNN_MtO(20, self.hidden_vector_size, 
                                            self.num_layers, self.n_classes, 'cpu')
        elif self.dtype == "residues":
            self.network = brnn_architecture.BRNN_MtM(20, self.hidden_vector_size, 
                                            self.num_layers, self.n_classes, 'cpu')
        else:
            raise ValueError("dtype must equal 'residues' or 'sequence'")
                                        
        self.network.load_state_dict(loaded_model)


    def predict(self, seq):
        """Use the network to predict values for a single sequence of valid amino acids

        Parameters
        ----------
        seq : str
            Valid amino acid sequence
            
        Returns
        -------
        np.ndarray
            Returns a 1D np.ndarray the length of the sequence where each position
            is the prediction at that position.
        """

        # convert sequence to uppercase
        seq = seq.upper()

        # Convert to one-hot sequence vector
        seq_vector = encode_sequence.one_hot(seq)
        seq_vector = seq_vector.view(1, len(seq_vector), -1)  # formatting

        # Forward pass
        prediction = self.network(seq_vector.float()).detach().numpy().flatten()

        # softmax to get class probabilities
        if self.task == "classification":
            if self.dtype == "residues":
                prediction = prediction.reshape(-1, self.n_classes)
                prediction = np.array(list(map(softmax, prediction)))

            else:
                prediction = softmax(prediction)

        return prediction



"""
Core training module of PARROT

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
import numpy as np

from parrot import brnn_plot
from parrot import encode_sequence


def train(network, train_loader, val_loader, datatype, problem_type, weights_file,
          stop_condition, device, learn_rate, n_epochs, verbose=False, silent=False):
    """Train a BRNN and save the best performing network weights

    Train the network on a training set, and every epoch evaluate its performance on
    a validation set. Save the network weights that acheive the best performance on
    the validation set.

    User must specify the machine learning tast (`problem_type`) and the format of
    the data (`datatype`). Additionally, this function requires the learning rate
    hyperparameter and the number of epochs of training. The other hyperparameters, 
    number of hidden layers and hidden vector size, are implictly included on the 
    the provided network.

    The user may specify if they want to train the network for a set number of
    epochs or until an automatic stopping condition is reached with the argument
    `stop_condition`. Depending on the stopping condition used, the `n_epochs`
    argument will have a different role.

    Parameters
    ----------
    network : PyTorch network object
            A BRNN network with the desired architecture
    train_loader : PyTorch DataLoader object
            A DataLoader containing the sequences and targets of the training set
    val_loader : PyTorch DataLoader object
            A DataLoader containing the sequences and targets of the validation set
    datatype : str
            The format of values in the dataset. Should be 'sequence' for datasets
            with a single value (or class label) per sequence, or 'residues' for
            datasets with values (or class labels) for every residue in a sequence.
    problem_type : str
            The machine learning task--should be either 'regression' or
            'classification'.
    weights_file : str
            A path to the location where the best_performing network weights will be
            saved
    stop_condition : str
            Determines when to conclude network training. If 'iter', then the network
            will train for `n_epochs` epochs, then stop. If 'auto' then the network
            will train for at least `n_epochs` epochs, then begin assessing whether
            performance has sufficiently stagnated. If the performance plateaus for
            `n_epochs` consecutive epochs, then training will stop.
    device : str
            Location of where training will take place--should be either 'cpu' or
            'cuda' (GPU). If available, training on GPU is typically much faster.
    learn_rate : float
            Initial learning rate of network training. The training process is
            controlled by the Adam optimization algorithm, so this learning rate
            will tend to decrease as training progresses.
    n_epochs : int
            Number of epochs to train for, or required to have stagnated performance
            for, depending on `stop_condition`.
    verbose : bool, optional
            If true, causes training updates to be written every epoch, rather than
            every 5 epochs.
    silent : bool, optional
            If true, causes not training updates to be written to standard out.

    Returns
    -------
    list
            A list of the average training set losses achieved at each epoch
    list
            A list of the average validation set losses achieved at each epoch
    """

    # Set optimizer
    optimizer = torch.optim.Adam(network.parameters(), lr=learn_rate)

    # Set loss criteria
    if problem_type == 'regression':
        if datatype == 'residues':
            criterion = nn.MSELoss(reduction='sum')
        elif datatype == 'sequence':
            criterion = nn.L1Loss(reduction='sum')
    elif problem_type == 'classification':
        criterion = nn.CrossEntropyLoss(reduction='sum')

    network = network.float()
    total_step = len(train_loader)
    min_val_loss = np.inf
    avg_train_losses = []
    avg_val_losses = []

    if stop_condition == 'auto':
        min_epochs = n_epochs
        # Set to some arbitrarily large number of iterations -- will stop automatically
        n_epochs = 20000000
        last_decrease = 0

    # Train the model - evaluate performance on val set every epoch
    end_training = False
    for epoch in range(n_epochs):  # Main loop

        # Initialize training and testing loss for epoch
        train_loss = 0
        val_loss = 0

        # Iterate over batches
        for i, (names, vectors, targets) in enumerate(train_loader):
            vectors = vectors.to(device)
            targets = targets.to(device)

            # Forward pass
            outputs = network(vectors.float())

            if problem_type == 'regression':
                loss = criterion(outputs, targets.float())
            else:
                if datatype == 'residues':
                    outputs = outputs.permute(0, 2, 1)
                loss = criterion(outputs, targets.long())

            train_loss += loss.data.item()

            # Backward and optimize
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        for names, vectors, targets in val_loader:
            vectors = vectors.to(device)
            targets = targets.to(device)

            # Forward pass
            outputs = network(vectors.float())
            if problem_type == 'regression':
                loss = criterion(outputs, targets.float())
            else:
                if datatype == 'residues':
                    outputs = outputs.permute(0, 2, 1)
                loss = criterion(outputs, targets.long())

            # Increment val loss
            val_loss += loss.data.item()

        # Avg loss:
        train_loss /= len(train_loader.dataset)
        val_loss /= len(val_loader.dataset)

        signif_decrease = True
        if stop_condition == 'auto' and epoch > min_epochs - 1:
            # Check to see if loss has stopped decreasing
            last_epochs_loss = avg_val_losses[-min_epochs:]

            for loss in last_epochs_loss:
                if val_loss >= loss*0.995:
                    signif_decrease = False

            # If network performance has plateaued over the last range of epochs, end training
            if not signif_decrease and epoch - last_decrease > min_epochs:
                end_training = True

        # Only save updated weights to memory if they improve val set performance
        if val_loss < min_val_loss:
            min_val_loss = val_loss 	# Reset min_val_loss
            last_decrease = epoch
            torch.save(network.state_dict(), weights_file)  # Save model

        # Append losses to lists
        avg_train_losses.append(train_loss)
        avg_val_losses.append(val_loss)

        if verbose:
            print('Epoch %d\tLoss %.4f' % (epoch, val_loss))
        elif epoch % 5 == 0 and silent is False:
            print('Epoch %d\tLoss %.4f' % (epoch, val_loss))

        # This is placed here to ensure that the best network, even if the performance
        # improvement is marginal, is saved.
        if end_training:
            break

    # Return loss per epoch so that they can be plotted
    return avg_train_losses, avg_val_losses


def test_labeled_data(network, test_loader, datatype,
                      problem_type, weights_file, num_classes,
                      probabilistic_classification, include_figs, 
                      device, output_file_prefix=''):
    """Test a trained BRNN on labeled sequences

    Using the saved weights of a trained network, run a set of sequences through
    the network and evaluate the performancd. Return the average loss per
    sequence and plot the results. Testing a network on previously-unseen data 
    provides a useful estimate of how generalizeable the network's performance is.

    Parameters
    ----------
    network : PyTorch network object
            A BRNN network with the desired architecture
    test_loader : PyTorch DataLoader object
            A DataLoader containing the sequences and targets of the test set
    datatype : str
            The format of values in the dataset. Should be 'sequence' for datasets
            with a single value (or class label) per sequence, or 'residues' for
            datasets with values (or class labels) for every residue in a sequence.
    problem_type : str
            The machine learning task--should be either 'regression' or
            'classification'.
    weights_file : str
            A path to the location of the best_performing network weights
    num_classes: int
            Number of data classes. If regression task, put 1.
    probabilistic_classification: bool
            Whether output should be binary labels, or "weights" of each label type.
            This field is only implemented for binary, sequence classification tasks.
    include_figs: bool
            Whether or not matplotlib figures should be generated.
    device : str
            Location of where testing will take place--should be either 'cpu' or
            'cuda' (GPU). If available, training on GPU is typically much faster.
    output_file_prefix : str
            Path and filename prefix to which the test set predictions and plots will be saved. 

    Returns
    -------
    float
            The average loss across the entire test set
    list of lists
            Details of the output predictions for each of the sequences in the test set. Each
            inner list represents a sample in the test set, with the format: [sequence_vector,
            true_value, predicted_value, sequence_ID]
    """

    # Load network weights
    network.load_state_dict(torch.load(weights_file))

    # Get output directory for images
    network_filename = weights_file.split('/')[-1]
    output_dir = weights_file[:-len(network_filename)]

    # Set loss criteria
    if problem_type == 'regression':
        criterion = nn.MSELoss()
    elif problem_type == 'classification':
        criterion = nn.CrossEntropyLoss()

    test_loss = 0
    all_targets = []
    all_outputs = []
    predictions = []
    for names, vectors, targets in test_loader: 	# batch size of 1
        all_targets.append(targets)

        vectors = vectors.to(device)
        targets = targets.to(device)

        # Forward pass
        outputs = network(vectors.float())
        if problem_type == 'regression':
            loss = criterion(outputs, targets.float())
        else:
            if datatype == 'residues':
                outputs = outputs.permute(0, 2, 1)
            loss = criterion(outputs, targets.long())

        test_loss += loss.data.item()  # Increment test loss
        all_outputs.append(outputs.detach())

        # Add to list as: [seq_vector, true value, predicted value, name]
        predictions.append([vectors[0].cpu().numpy(), targets.cpu().numpy()
                           [0], outputs.cpu().detach().numpy(), names[0]])

    # Plot 'accuracy' depending on the problem type and datatype
    if problem_type == 'regression':
        if datatype == 'residues':
            if include_figs:
                brnn_plot.residue_regression_scatterplot(all_targets, all_outputs, 
                                            output_file_prefix=output_file_prefix)

            # Format predictions
            for i in range(len(predictions)):
                predictions[i][2] = predictions[i][2].flatten()
                predictions[i][1] = predictions[i][1].flatten()

        elif datatype == 'sequence':
            if include_figs:
                brnn_plot.sequence_regression_scatterplot(all_targets, all_outputs, 
                                            output_file_prefix=output_file_prefix)

            # Format predictions
            for i in range(len(predictions)):
                predictions[i][2] = predictions[i][2][0][0]
                predictions[i][1] = predictions[i][1][0]

    elif problem_type == 'classification':

        if datatype == 'residues':
            if include_figs:
                brnn_plot.res_confusion_matrix(all_targets, all_outputs, num_classes, 
                                            output_file_prefix=output_file_prefix)

            # Format predictions and assign class predictions
            for i in range(len(predictions)):
                pred_values = []
                for j in range(len(predictions[i][2])):
                    pred_values = np.argmax(predictions[i][2], axis=1)[0]
                predictions[i][2] = np.array(pred_values, dtype=np.int)

        elif datatype == 'sequence':
            if probabilistic_classification:
                # Probabilistic assignment of class predictions
                # Optional implementation for classification tasks
                # e.g. every sequence is assigned probabilities
                # corresponding to each possible class
                pred_probabilities = []
                for i in range(len(predictions)):
                    softmax = np.exp(predictions[i][2][0])
                    probs = softmax / np.sum(softmax)
                    predictions[i][2] = probs
                    pred_probabilities.append(probs)

                # Plot ROC and PR curves
                if include_figs:
                    brnn_plot.plot_roc_curve(all_targets, pred_probabilities, num_classes, 
                                            output_file_prefix=output_file_prefix)
                    brnn_plot.plot_precision_recall_curve(all_targets, pred_probabilities, 
                                            num_classes, output_file_prefix=output_file_prefix)

            else:
                # Absolute assignment of class predictions
                # e.g. every sequence receives an integer class label
                for i in range(len(predictions)):
                    pred_value = np.argmax(predictions[i][2])
                    predictions[i][2] = int(pred_value)

                # Plot confusion matrix (if not in probabilistic classification mode)
                if include_figs:
                    brnn_plot.confusion_matrix(all_targets, all_outputs, num_classes, 
                                                output_file_prefix=output_file_prefix)

    return test_loss / len(test_loader.dataset), predictions


def test_unlabeled_data(network, sequences, device, encoding_scheme='onehot', encoder=None, print_frequency=None):
    """Test a trained BRNN on unlabeled sequences

    Use a trained network to make predictions on previously-unseen data.

    ** 
    Note: Unlike the previous functions, `network` here must have pre-loaded
    weights. 
    **

    Parameters
    ----------
    network : PyTorch network object
            A BRNN network with the desired architecture and pre-loaded weights
    sequences : list
            A list of amino acid sequences to test using the network
    device : str
            Location of where testing will take place--should be either 'cpu' or
            'cuda' (GPU). If available, training on GPU is typically much faster.
    encoding_scheme : str, optional
            How amino acid sequences are to be encoded as numeric vectors. Currently,
            'onehot','biophysics' and 'user' are the implemented options.
    encoder: UserEncoder object, optional
            If encoding_scheme is 'user', encoder should be a UserEncoder object
            that can convert amino acid sequences to numeric vectors. If
            encoding_scheme is not 'user', use None.
    print_frequency : int
            If provided defines at what sequence interval an update is printed.
            Default = None.
    
    Returns
    -------
    dict
            A dictionary containing predictions mapped to sequences
    """

    pred_dict = {}

    local_count = -1
    total_count = len(sequences)

    for seq in sequences:

        local_count = local_count + 1
        if print_frequency is not None:
            if local_count % print_frequency == 0:
                print(f'On {local_count} of {total_count}')

        if encoding_scheme == 'onehot':
            seq_vector = encode_sequence.one_hot(seq)
        elif encoding_scheme == 'biophysics':
            seq_vector = encode_sequence.biophysics(seq)
        elif encoding_scheme == 'user':
            seq_vector = encoder.encode(seq)

        seq_vector = seq_vector.view(1, len(seq_vector), -1)

        # Forward pass
        outputs = network(seq_vector.float()).detach().numpy()
        pred_dict[seq] = outputs

    return pred_dict



"""
Plot training results for regression and classification tasks on both 
sequence-mapped and residue-mapped data.

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import numpy as np
import torch
import itertools
from scipy.stats import linregress, pearsonr, spearmanr
from sklearn.metrics import roc_curve, auc
from sklearn.metrics import precision_recall_curve, average_precision_score
from sklearn.metrics import f1_score, matthews_corrcoef, accuracy_score
import matplotlib.pyplot as plt
import seaborn as sn
import pandas as pd

from parrot import encode_sequence


def training_loss(train_loss, val_loss, output_file_prefix=''):
    """Plot training and validation loss per epoch

    Figure is saved to file at "<output_file_prefix>_train_val_loss.png".

    Parameters
    ----------
    train_loss : list
            training loss across each epoch
    val_loss : list
            validation loss across each epoch
    output_file_prefix : str, optional
            File to which the plot will be saved as "<output_file_prefix>_train_val_loss.png"
    """

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6, 6))
    props = dict(boxstyle='round', facecolor='gainsboro', alpha=0.5)

    num_epochs = len(train_loss)

    # Loss per epoch
    training_loss, = ax.plot(np.arange(1, num_epochs+1), train_loss, label='Train')
    validation_loss, = ax.plot(np.arange(1, num_epochs+1), val_loss, label='Val')
    ax.set_xlabel("Epoch")
    ax.set_ylabel("Avg loss")
    ax.set_title("Training and testing loss per epoch")
    ax.legend(handles=[training_loss, validation_loss], fontsize=14,
              facecolor='gainsboro', edgecolor='slategray')

    if num_epochs < 21:
        ax.set_xticks(np.arange(2, num_epochs+1, 2))
    elif num_epochs < 66:
        ax.set_xticks(np.arange(5, num_epochs+1, 5))
    elif num_epochs < 151:
        ax.set_xticks(np.arange(10, num_epochs+1, 10))
    else:
        ax.set_xticks(np.arange(50, num_epochs+1, 50))

    plt.savefig(output_file_prefix + '_train_val_loss.png')
    plt.clf()


def sequence_regression_scatterplot(true, predicted, output_file_prefix=''):
    """Create a scatterplot for a sequence-mapped values regression problem

    Figure is saved to file at "<output_file_prefix>_seq_scatterplot.png".

    Parameters
    ----------
    true : list of PyTorch FloatTensors
            A list where each item is a [1 x 1] tensor with the true regression value
            of a particular sequence
    predicted : list of PyTorch FloatTensors
            A list where each item is a [1 x 1] tensor with the regression prediction
            for a particular sequence
    output_file_prefix : str, optional
            File to which the plot will be saved as "<output_file_prefix>_seq_scatterplot.png"
    """

    true_list = []
    pred_list = []

    for item in true:
        true_list.append(item.cpu().numpy()[0][0])
    for item in predicted:
        pred_list.append(item.cpu().numpy()[0][0])

    plt.scatter(true_list, pred_list)
    edge_vals = [0.9*min(min(true_list), min(pred_list)),
                 1.1*max(max(true_list), max(pred_list))]
    plt.xlim(edge_vals)
    plt.ylim(edge_vals)
    plt.plot(edge_vals, edge_vals, 'k--')
    plt.xlabel('True')
    plt.ylabel('Predicted')
    slope, intercept, r_value, p_value, std_err = linregress(true_list, pred_list)
    plt.title('Testing accuracy: R^2=%.3f' % (r_value**2))
    plt.savefig(output_file_prefix + '_seq_scatterplot.png')


def residue_regression_scatterplot(true, predicted, output_file_prefix=''):
    """Create a scatterplot for a residue-mapped values regression problem

    Each sequence is plotted with a unique marker-color combination, up to 70
    different sequences.

    Figure is saved to file at "<output_file_prefix>_res_scatterplot.png".

    Parameters
    ----------
    true : list of PyTorch FloatTensors
            A list where each item is a [1 x len(sequence)] tensor with the true
            regression values of each residue in a sequence
    predicted : list of PyTorch FloatTensors
            A list where each item is a [1 x len(sequence)] tensor with the 
            regression predictions for each residue in a sequence
    output_file_prefix : str, optional
            File to which the plot will be saved as "<output_file_prefix>_res_scatterplot.png"
    """

    true_list = []
    pred_list = []

    marker = itertools.cycle(('>', '+', '.', 'o', '*', 'v', 'D'))

    for item in true:
        single_frag = item.cpu().numpy()[0].flatten()
        true_list.append(list(single_frag))
    for item in predicted:
        single_frag = item.cpu().numpy()[0].flatten()
        pred_list.append(list(single_frag))

    for i in range(len(true_list)):
        plt.scatter(true_list[i], pred_list[i], s=6, marker=next(marker))

    plt.figure(1)

    left, right = plt.xlim()
    bottom, top = plt.ylim()
    edge_vals = [min(left, bottom), max(right, top)]
    plt.xlim(edge_vals)
    plt.ylim(edge_vals)
    plt.plot(edge_vals, edge_vals, 'k--')
    plt.xlabel('True')
    plt.ylabel('Predicted')
    slope, intercept, r_value, p_value, std_err = linregress(sum(true_list, []), sum(pred_list, []))
    plt.title('Testing accuracy: R^2=%.3f' % (r_value**2))
    plt.savefig(output_file_prefix + '_res_scatterplot.png')

def plot_roc_curve(true_classes, predicted_class_probs, num_classes, output_file_prefix=''):
    """Create an ROC curve for a sequence classification problem

    Figure is saved to file at "<output_file_prefix>_ROC_curve.png".

    Parameters
    ----------
    true_classes : list of PyTorch IntTensors
            A list where each item is a [1 x 1] tensor with the true class label of a
            particular sequence
    predicted_class_probs : list of PyTorch FloatTensors
            A list where each item is a [1 x num_classes] tensor of the probabilities
            of assignment to each class
    num_classes : int
            Number of distinct data classes
    output_file_prefix : str, optional
            File to which the plot will be saved as "<output_file_prefix>_ROC_curve.png"
    """

    y_test = np.zeros((len(true_classes), num_classes), dtype=int)
    for i in range(len(true_classes)):
        label = true_classes[i].numpy()[0]
        y_test[i, label] = 1
    y_score = np.vstack(predicted_class_probs)

    # Compute ROC curve and ROC area for each class
    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    for c in range(num_classes):
        fpr[c], tpr[c], _ = roc_curve(y_test[:, c], y_score[:, c])
        roc_auc[c] = auc(fpr[c], tpr[c])

    plt.figure()
    if num_classes > 2:
        # Compute micro-average ROC curve and ROC area (if multiclass)
        fpr["micro"], tpr["micro"], _ = roc_curve(y_test.ravel(), y_score.ravel())
        roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])

        # Plot all ROC curves
        plt.plot(fpr["micro"], tpr["micro"],
                 label='Average (area = {0:0.2f})'
                       ''.format(roc_auc["micro"]),
                 color='deeppink', linestyle=':', linewidth=4)

        for c in range(num_classes):
            plt.plot(fpr[c], tpr[c], lw=2,
                     label='Class {0} (area = {1:0.2f})'
                     ''.format(c, roc_auc[c]))

    elif num_classes == 2: # If binary classification
        # Plot only one curve (doesn't matter which one, they are symmetric)
        plt.plot(fpr[1], tpr[1], lw=2,
                     label='Binary class (area = {0:0.2f})'
                     ''.format(roc_auc[1]))

    plt.plot([0, 1], [0, 1], 'k--', lw=2)
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver operating characteristic')
    plt.legend(loc="lower right", fontsize=8)
    plt.savefig(output_file_prefix + '_ROC_curve.png')

def plot_precision_recall_curve(true_classes, predicted_class_probs, 
                                num_classes, output_file_prefix=''):
    """Create an PR curve for a sequence classification problem

    Figure is saved to file at "<output_file_prefix>_PR_curve.png".

    Parameters
    ----------
    true_classes : list of PyTorch IntTensors
            A list where each item is a [1 x 1] tensor with the true class label of a
            particular sequence
    predicted_class_probs : list of PyTorch FloatTensors
            A list where each item is a [1 x num_classes] tensor of the probabilities
            of assignment to each class
    num_classes : int
            Number of distinct data classes
    output_file_prefix : str, optional
            File to which the plot will be saved as "<output_file_prefix>_PR_curve.png"
    """

    y_test = np.zeros((len(true_classes), num_classes), dtype=int)
    for i in range(len(true_classes)):
        label = true_classes[i].numpy()[0]
        y_test[i, label] = 1
    y_score = np.vstack(predicted_class_probs)

    # For each class
    precision = dict()
    recall = dict()
    average_precision = dict()
    for i in range(num_classes):
        precision[i], recall[i], _ = precision_recall_curve(y_test[:, i],
                                                            y_score[:, i])
        average_precision[i] = average_precision_score(y_test[:, i], y_score[:, i])

    # A "micro-average": quantifying score on all classes jointly
    precision["micro"], recall["micro"], _ = precision_recall_curve(y_test.ravel(),
        y_score.ravel())
    average_precision["micro"] = average_precision_score(y_test, y_score,
                                                        average="micro")

    # Plot
    plt.figure()
    plt.plot(recall["micro"], precision["micro"], color='deeppink', linestyle=':', 
        linewidth=4, label='Average (area = {0:0.2f})'
              ''.format(average_precision["micro"]))
    for c in range(num_classes):
        plt.plot(recall[c], precision[c], lw=2, 
            label='Class {0} (area = {1:0.2f})'
                  ''.format(c, average_precision[c]))

    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    plt.title('Precision-Recall')
    plt.legend()
    plt.savefig(output_file_prefix + '_PR_curve.png')


def confusion_matrix(true_classes, predicted_classes, num_classes, output_file_prefix=''):
    """Create a confusion matrix for a sequence classification problem

    Figure is saved to file at "<output_file_prefix>_seq_CM.png".

    Parameters
    ----------
    true_classes : list of PyTorch IntTensors
            A list where each item is a [1 x 1] tensor with the true class label of a
            particular sequence
    predicted_classes : list of PyTorch FloatTensors
            A list where each item is a [1 x num_classes] tensor prediction of the
            class label for a particular sequence
    num_classes : int
            Number of distinct data classes
    output_file_prefix : str, optional
            File to which the plot will be saved as "<output_file_prefix>_seq_CM.png"
    """

    cm = np.zeros((num_classes, num_classes))
    for i in range(len(true_classes)):
        cm[np.argmax(predicted_classes[i][0].cpu().numpy()), true_classes[i][0]] += 1

    df_cm = pd.DataFrame(cm, range(num_classes), range(num_classes))
    sn.set(font_scale=1.4)  # for label size
    sn.heatmap(df_cm, cmap='Blues', annot=True, annot_kws={"size": 16})  # font size
    plt.xlabel('True labels')
    plt.ylabel('Predicted labels')
    plt.title('Test set confusion matrix')
    plt.tight_layout()
    plt.savefig(output_file_prefix + '_seq_CM.png')


def res_confusion_matrix(true_classes, predicted_classes, num_classes, output_file_prefix=''):
    """Create a confusion matrix for a residue classification problem

    Figure is saved to file at "<output_file_prefix>_res_CM.png".

    Parameters
    ----------
    true_classes : list of PyTorch IntTensors
            A list where each item is a [1 x len(sequence)] tensor with the true class
            label of the residues in a particular sequence
    predicted_classes : list of PyTorch FloatTensors
            A list where each item is a [1 x num_classes x len(sequence)] tensor
            with predictions of the class label for each residue in a particular
            sequence
    num_classes : int
            Number of distinct data classes
    output_file_prefix : str, optional
            File to which the plot will be saved as "<output_file_prefix>_res_CM.png"
    """

    true_list = []
    pred_list = []

    for item in true_classes:
        single_frag = list(item[0].cpu().numpy().flatten())
        true_list = true_list + single_frag

    for item in predicted_classes:
        single_frag = item[0].permute(1, 0).cpu().numpy()

        for residue in single_frag:
            pred_list.append(np.argmax(residue))

    cm = np.zeros((num_classes, num_classes))
    for i in range(len(true_list)):
        cm[pred_list[i], true_list[i]] += 1

    df_cm = pd.DataFrame(cm, range(num_classes), range(num_classes))
    sn.set(font_scale=1.4)  # for label size
    sn.heatmap(df_cm, cmap='Blues', annot=True, annot_kws={"size": 16})  # font size
    plt.xlabel('True labels')
    plt.ylabel('Predicted labels')
    plt.title('Test set confusion matrix')
    plt.tight_layout()
    plt.savefig(output_file_prefix + '_res_CM.png')


def write_performance_metrics(sequence_data, dtype, problem_type,
                                prob_class, output_file_prefix=''):
    """Writes a short text file describing performance on a variety of metrics

    Writes different output depending on whether a classification or regression task
    is specified. Also produces unique output if in probabilistic classification mode.
    File is saved to "<output_file_prefix>_performance_stats.txt".

    Parameters
    ----------
    sequence_data : list of lists
            Details of the output predictions for each of the sequences in the test set. Each
            inner list represents a sample in the test set, with the format: [sequence_vector,
            true_value, predicted_value, sequence_ID]
    dtype : str
            The format of values in the dataset. Should be 'sequence' for datasets
            with a single value (or class label) per sequence, or 'residues' for
            datasets with values (or class labels) for every residue in a sequence.
    problem_type : str
            The machine learning task--should be either 'regression' or 'classification'.
    prob_class : bool
            Flag indicating if probabilistic classification was specified by the user.
    output_file_prefix : str
            Path and filename prefix to which the test set predictions will be saved. Final
            file path is "<output_file_prefix>_performance_stats.txt"
    """

    true_vals = [l[1] for l in sequence_data]
    pred_vals = [l[2] for l in sequence_data]

    perform_metrics = {}

    if dtype == 'residues':
        true_vals = np.hstack(true_vals)
        pred_vals = np.hstack(pred_vals)

    if problem_type == 'classification':
        # Take care of probabilistic-classification case first
        if prob_class:
            # Reformat
            pred_vals = np.vstack(pred_vals)
            true_vals_array = np.zeros((len(true_vals), len(pred_vals[0])), dtype=int)
            for i in range(len(true_vals)):
                true_vals_array[i, true_vals[i]] = 1

            # AUROC, AUPRC
            perform_metrics['Area under Precision-Recall curve'] = round(
                                        average_precision_score(true_vals_array, 
                                        pred_vals, average="micro"), 3)
            fpr, tpr, _ = roc_curve(true_vals_array.ravel(), pred_vals.ravel())
            perform_metrics["Area under ROC"] = round(auc(fpr, tpr), 3)

            # Change probs to discrete classes
            pred_vals = np.argmax(pred_vals, axis=1)

        # Then take care of general classification stats: accuracy, F1, MCC
        perform_metrics['Matthews Correlation Coef'] = round(
                            matthews_corrcoef(true_vals, pred_vals), 3)
        perform_metrics['F1 Score'] = round(
                            f1_score(true_vals, pred_vals, average='weighted'), 3)
        perform_metrics['Accuracy'] = round(accuracy_score(true_vals, pred_vals), 3)


    elif problem_type == 'regression':
        # Pearson R, Spearman R
        pears_r, p_val = pearsonr(true_vals, pred_vals)
        perform_metrics['Pearson R'] = round(pears_r, 3)
        spearman_r, p_val = spearmanr(true_vals, pred_vals)
        perform_metrics['Spearman R'] = round(spearman_r, 3)     

    # Write performance metrics to file
    with open(output_file_prefix + '_performance_stats.txt', 'w') as f:
        for key, value in perform_metrics.items():
            outstr = '%s : %.3f\n' % (key, value)
            f.write(outstr)


def output_predictions_to_file(sequence_data, excludeSeqID, encoding_scheme,
                            probabilistic_class, encoder=None, output_file_prefix=''):
    """Output sequences, their true values, and their predicted values to a file

    Used on the output of the test_unlabeled_data() function in the train_network module in
    order to detail the performance of the trained network on the test set. Produces the
    file "test_set_predictions.tsv" in output_dir. Each pair of lines in this tsvfile
    corresponds to a particular test set sequence, with the first containing the true data
    values, and the second line having the predicted data values.

    Parameters
    ----------
    sequence_data : list of lists
            Details of the output predictions for each of the sequences in the test set. Each
            inner list represents a sample in the test set, with the format: [sequence_vector,
            true_value, predicted_value, sequence_ID]
    excludeSeqID : bool
            Boolean indicating whether or not each line in `tsvfile` has a sequence ID
            (default is False)		
    encoding_scheme : str
            Description of how an amino acid sequence should be encoded as a numeric 
            vector. Providing a string other than 'onehot', 'biophysics', or 'user' 
            will produce unintended consequences.
    probabilistic_class : bool
            Flag indicating if probabilistic classification was specified by the user. If True,
            instead of class labels, predictions will be output as probabilities of each class.
    encoder: UserEncoder object, optional
            If encoding_scheme is 'user', encoder should be a UserEncoder object
            that can convert amino acid sequences to numeric vectors. If
            encoding_scheme is not 'user', use None.
    output_file_prefix : str
            Path and filename prefix to which the test set predictions will be saved. Final
            file path is "<output_file_prefix>_predictions.tsv"
    """

    seq_vectors = []
    true_vals = []
    pred_vals = []
    names = []
    count = 0
    for sequence in sequence_data:
        seq_vector, true_val, pred_val, name = sequence
        seq_vectors.append(seq_vector)
        true_vals.append(true_val)
        pred_vals.append(pred_val)

        if excludeSeqID:
            names.append('test' + str(count))
            count += 1
        else:
            names.append(name)

    # Decode the sequence vectors
    if encoding_scheme == 'onehot':
        sequences = encode_sequence.rev_one_hot(seq_vectors)
    elif encoding_scheme == 'biophysics':
        sequences = encode_sequence.rev_biophysics(seq_vectors)
    else:
        sequences = encoder.decode(seq_vectors)

    # Write to file
    with open(output_file_prefix + '_predictions.tsv', 'w') as tsvfile:
        for i in range(len(names)):

            # Adjust formatting for residues or sequence data
            if isinstance(true_vals[i], np.ndarray):
                true_vals_format = ' '.join(true_vals[i].astype(str))
                pred_vals_format = ' '.join(pred_vals[i].astype(str))
            elif probabilistic_class:
                true_vals_format = true_vals[i]
                pred_vals_format = ' '.join(np.around(pred_vals[i], decimals=4).astype(str))
            else:
                true_vals_format = true_vals[i]
                pred_vals_format = pred_vals[i]

            '''
			Format:
			NAME_TRUE SEQUENCE TRUE_VALUE(S)
			NAME_PRED SEQUENCE PRED_VALUE(S)
			'''
            output_str = "%s_TRUE %s %s\n" % (names[i], sequences[i], true_vals_format)
            output_str = output_str + "%s_PRED %s %s\n" % (names[i], sequences[i], pred_vals_format)

            tsvfile.write(output_str)



"""
PARROT
A bidirectional recurrent neural network framework for protein bioinformatics
"""

# Add imports here
from .encode_sequence import *
from .process_input_data import *
from .brnn_architecture import *
from .brnn_plot import *
from .train_network import *
from .bayesian_optimization import *

# Handle versioneer
from ._version import get_versions
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions

# code that allows access to the data directory
_ROOT = os.path.abspath(os.path.dirname(__file__))
def get_directory(path='../data'):
    return os.path.join(_ROOT, path)



"""
PARROT
A bidirectional recurrent neural network framework for protein bioinformatics
"""
import sys
from setuptools import setup, find_packages
import versioneer

short_description = __doc__.split("\n")

# from https://github.com/pytest-dev/pytest-runner#conditional-requirement
needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

try:
    with open("README.md", "r") as handle:
        long_description = handle.read()
except:
    long_description = "\n".join(short_description[2:])


setup(
    # Self-descriptive entries which should always be present
    name='idptools-parrot',
    author='Holehouse Lab',
    author_email='degriffith@wustl.edu',
    description=short_description[0],
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license='MIT',

    # Which Python importable modules should be included when your package is installed
    # Handled automatically by setuptools. Use 'exclude' to prevent some specific
    # subpackage(s) from being added, if needed
    packages=find_packages(),

    # Optional include package data to ship with your package
    # Customize MANIFEST.in if the general case does not suit your needs
    # Comment out this line to prevent the files from being packaged with your software
    # include_package_data=True,

    # Include data files
    # package_data={},

    # Allows `setup.py test` to work correctly with pytest
    setup_requires=[] + pytest_runner,
    scripts=['scripts/parrot-train',
             'scripts/parrot-predict',
             'scripts/parrot-optimize',
             'scripts/parrot-cvsplit',
             'scripts/parrot-preprocess'],

    # Required packages, pulls from pip if needed; do not use for Conda deployment
    install_requires=[
            'cython',
            'torch>=1.8.0',
            'numpy',
            'argparse',
            'more-itertools',
            'scipy',
            'scikit-learn',
            'matplotlib',
            'seaborn',
            'pandas'],

    extras_require={
            'optimize': ['GPy', 'GPyOpt'] },


    python_requires=">=3.9,<3.10.0",          # Python version restrictions

    # Manual control if final package is compressible or not, set False to prevent the .egg from being made
    zip_safe=False,

)



"""
The underlying architecture of the bidirectional LSTM network used in PARROT

.............................................................................
idptools-parrot was developed by the Holehouse lab
     Original release ---- 2020

Question/comments/concerns? Raise an issue on github:
https://github.com/idptools/parrot

Licensed under the MIT license. 
"""

import torch
import torch.nn as nn


class BRNN_MtM(nn.Module):
    """A PyTorch many-to-many bidirectional recurrent neural network

    A class containing the PyTorch implementation of a BRNN. The network consists
    of repeating LSTM units in the hidden layers that propogate sequence information
    in both the foward and reverse directions. A final fully connected layer
    aggregates the deepest hidden layers of both directions and produces the
    outputs.

    "Many-to-many" refers to the fact that the network will produce outputs 
    corresponding to every item of the input sequence. For example, an input 
    sequence of length 10 will produce 10 sequential outputs.

    Attributes
    ----------
    device : str
        String describing where the network is physically stored on the computer.
        Should be either 'cpu' or 'cuda' (GPU).
    hidden_size : int
        Size of hidden vectors in the network
    num_layers : int
        Number of hidden layers (for each direction) in the network
    num_classes : int
        Number of classes for the machine learning task. If it is a regression
        problem, `num_classes` should be 1. If it is a classification problem,
        it should be the number of classes.
    lstm : PyTorch LSTM object
        The bidirectional LSTM layer(s) of the recurrent neural network.
    fc : PyTorch Linear object  
        The fully connected linear layer of the recurrent neural network. Across 
        the length of the input sequence, this layer aggregates the output of the
        LSTM nodes from the deepest forward layer and deepest reverse layer and
        returns the output for that residue in the sequence.
    """

    def __init__(self, input_size, hidden_size, num_layers, num_classes, device):
        """
        Parameters
        ----------
        input_size : int
            Length of the input vectors at each timestep
        hidden_size : int
            Size of hidden vectors in the network
        num_layers : int
            Number of hidden layers (for each direction) in the network
        num_classes : int
            Number of classes for the machine learning task. If it is a regression
            problem, `num_classes` should be 1. If it is a classification problem,
            it should be the number of classes.
        device : str
            String describing where the network is physically stored on the computer.
            Should be either 'cpu' or 'cuda' (GPU).
        """

        super(BRNN_MtM, self).__init__()
        self.device = device
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.num_classes = num_classes
        self.lstm = nn.LSTM(input_size, hidden_size, num_layers,
                            batch_first=True, bidirectional=True)
        self.fc = nn.Linear(in_features=hidden_size*2,  # *2 for bidirection
                            out_features=num_classes)

    def forward(self, x):
        """Propogate input sequences through the network to produce outputs

        Parameters
        ----------
        x : 3-dimensional PyTorch IntTensor
            Input sequence to the network. Should be in the format:
            [batch_dim X sequence_length X input_size]

        Returns
        -------
        3-dimensional PyTorch FloatTensor
            Output after propogating the sequences through the network. Will
            be in the format:
            [batch_dim X sequence_length X num_classes]
        """

        # Set initial states
        # h0 and c0 dimensions: [num_layers*2 X batch_size X hidden_size]
        h0 = torch.zeros(self.num_layers*2,     # *2 for bidirection
                         x.size(0), self.hidden_size).to(self.device)
        c0 = torch.zeros(self.num_layers*2,
                         x.size(0), self.hidden_size).to(self.device)

        # Forward propagate LSTM
        # out: tensor of shape: [batch_size, seq_length, hidden_size*2]
        out, (h_n, c_n) = self.lstm(x, (h0, c0))

        # Decode the hidden state for each time step
        fc_out = self.fc(out)
        return fc_out


class BRNN_MtO(nn.Module):
    """A PyTorch many-to-one bidirectional recurrent neural network

    A class containing the PyTorch implementation of a BRNN. The network consists
    of repeating LSTM units in the hidden layers that propogate sequence information
    in both the foward and reverse directions. A final fully connected layer
    aggregates the deepest hidden layers of both directions and produces the
    output.

    "Many-to-one" refers to the fact that the network will produce a single output 
    for an entire input sequence. For example, an input sequence of length 10 will
    produce only one output.

    Attributes
    ----------
    device : str
        String describing where the network is physically stored on the computer.
        Should be either 'cpu' or 'cuda' (GPU).
    hidden_size : int
        Size of hidden vectors in the network
    num_layers : int
        Number of hidden layers (for each direction) in the network
    num_classes : int
        Number of classes for the machine learning task. If it is a regression
        problem, `num_classes` should be 1. If it is a classification problem,
        it should be the number of classes.
    lstm : PyTorch LSTM object
        The bidirectional LSTM layer(s) of the recurrent neural network.
    fc : PyTorch Linear object  
        The fully connected linear layer of the recurrent neural network. Across 
        the length of the input sequence, this layer aggregates the output of the
        LSTM nodes from the deepest forward layer and deepest reverse layer and
        returns the output for that residue in the sequence.
    """

    def __init__(self, input_size, hidden_size, num_layers, num_classes, device):
        """
        Parameters
        ----------
        input_size : int
            Length of the input vectors at each timestep
        hidden_size : int
            Size of hidden vectors in the network
        num_layers : int
            Number of hidden layers (for each direction) in the network
        num_classes : int
            Number of classes for the machine learning task. If it is a regression
            problem, `num_classes` should be 1. If it is a classification problem,
            it should be the number of classes.
        device : str
            String describing where the network is physically stored on the computer.
            Should be either 'cpu' or 'cuda' (GPU).
        """

        super(BRNN_MtO, self).__init__()
        self.device = device
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.lstm = nn.LSTM(input_size, hidden_size, num_layers,
                            batch_first=True, bidirectional=True)
        self.fc = nn.Linear(in_features=hidden_size*2,  # *2 for bidirection
                            out_features=num_classes)

    def forward(self, x):
        """Propogate input sequences through the network to produce outputs

        Parameters
        ----------
        x : 3-dimensional PyTorch IntTensor
            Input sequence to the network. Should be in the format:
            [batch_dim X sequence_length X input_size]

        Returns
        -------
        3-dimensional PyTorch FloatTensor
            Output after propogating the sequences through the network. Will
            be in the format:
            [batch_dim X 1 X num_classes]
        """

        # Set initial states
        # h0 and c0 dimensions: [num_layers*2 X batch_size X hidden_size]
        h0 = torch.zeros(self.num_layers*2,     # *2 for bidirection
                         x.size(0), self.hidden_size).to(self.device)
        c0 = torch.zeros(self.num_layers*2,
                         x.size(0), self.hidden_size).to(self.device)

        # Forward propagate LSTM
        # out: tensor of shape: [batch_size, seq_length, hidden_size*2]
        out, (h_n, c_n) = self.lstm(x, (h0, c0))

        # Retain the outputs of the last time step in the sequence for both directions
        # (i.e. output of seq[n] in forward direction, seq[0] in reverse direction)
        final_outs = torch.cat((h_n[:, :, :][-2, :], h_n[:, :, :][-1, :]), -1)

        # Decode the hidden state of the last time step
        fc_out = self.fc(final_outs)
        return fc_out
